#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .. import core as c
from .. import utils as ut
from ..localize import _
from .basicstru import EUDJump, EUDJumpIf, EUDJumpIfNot
from .cshelper import CtrlStruOpener


def _IsLoopBlock(block):
    return "contpoint" in block


# -------


def EUDInfLoop():
    def _footer():
        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
        }

        ut.EUDCreateBlock("infloopblock", block)
        return True

    return CtrlStruOpener(_footer)


def EUDEndInfLoop():
    block = ut.EUDPopBlock("infloopblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << block["loopstart"]
    EUDJump(block["loopstart"])
    block["loopend"] << c.NextTrigger()


# -------


def EUDLoopN():
    def _footer(n):
        vardb = c.Db(4)
        c.RawTrigger(actions=c.SetMemory(vardb, c.SetTo, n))

        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
            "vardb": vardb,
        }

        ut.EUDCreateBlock("loopnblock", block)
        EUDJumpIf(c.Memory(vardb, c.Exactly, 0), block["loopend"])
        return True

    return CtrlStruOpener(_footer)


def EUDEndLoopN():
    block = ut.EUDPopBlock("loopnblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << c.NextTrigger()

    vardb = block["vardb"]
    c.RawTrigger(actions=c.SetMemory(vardb, c.Subtract, 1))
    EUDJump(block["loopstart"])
    block["loopend"] << c.NextTrigger()


# -------


def EUDLoopRange(start, end=None):
    ut.EUDCreateBlock("looprangeblock", None)
    if end is None:
        start, end = 0, start

    v = c.EUDVariable()
    v << start
    if EUDWhile()(v < end):
        yield v
        EUDLoopSetContinuePoint()
        v += 1
    EUDEndWhile()
    ut.EUDPopBlock("looprangeblock")


# -------


def EUDWhile():
    def _header():
        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
        }

        ut.EUDCreateBlock("whileblock", block)

    def _footer(conditions, *, neg=False):
        block = ut.EUDPeekBlock("whileblock")[1]
        if neg:
            EUDJumpIf(conditions, block["loopend"])
        else:
            EUDJumpIfNot(conditions, block["loopend"])
        return True

    _header()
    return CtrlStruOpener(_footer)


def EUDWhileNot():
    c = EUDWhile()
    return CtrlStruOpener(lambda conditions: c(conditions, neg=True))


def EUDEndWhile():
    block = ut.EUDPopBlock("whileblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << block["loopstart"]
    EUDJump(block["loopstart"])
    block["loopend"] << c.NextTrigger()


# -------


def _GetLastLoopBlock():
    for block in reversed(ut.EUDGetBlockList()):
        if _IsLoopBlock(block[1]):
            return block

    raise ut.EPError(_("No loop block surrounding this code area"))


def EUDLoopContinue():
    block = _GetLastLoopBlock()[1]
    EUDJump(block["contpoint"])


def EUDLoopContinueIf(conditions):
    block = _GetLastLoopBlock()[1]
    EUDJumpIf(conditions, block["contpoint"])


def EUDLoopContinueIfNot(conditions):
    block = _GetLastLoopBlock()[1]
    EUDJumpIfNot(conditions, block["contpoint"])


def EUDLoopIsContinuePointSet():
    block = _GetLastLoopBlock()[1]
    return block["contpoint"].IsSet()


def EUDLoopSetContinuePoint():
    block = _GetLastLoopBlock()[1]
    block["contpoint"] << c.NextTrigger()


def EUDLoopBreak():
    block = _GetLastLoopBlock()[1]
    EUDJump(block["loopend"])


def EUDLoopBreakIf(conditions):
    block = _GetLastLoopBlock()[1]
    EUDJumpIf(conditions, block["loopend"])


def EUDLoopBreakIfNot(conditions):
    block = _GetLastLoopBlock()[1]
    EUDJumpIfNot(conditions, block["loopend"])



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from ... import core as c
from ... import ctrlstru as cs
from ... import eudlib as sf
from ... import utils as ut
from ...trigtrg.runtrigtrg import _runner_cp

sharedTriggers = []


def GetExecutingPlayers(bTrigger):
    # Get executing players of the trigger.
    # If all player executes it, then pass it
    if bTrigger[320 + 2048 + 4 + 17] != 0:
        playerExecutesTrigger = [True] * 8

    else:  # Should check manually
        playerExecutesTrigger = [False] * 8
        # By player
        for player in range(8):
            if bTrigger[320 + 2048 + 4 + player] != 0:
                playerExecutesTrigger[player] = True

        # By force
        playerForce = [0] * 8
        for player in range(8):
            playerForce[player] = c.GetPlayerInfo(player).force

        for force in range(4):
            if bTrigger[320 + 2048 + 4 + 18 + force] != 0:
                for player in range(8):
                    if playerForce[player] == force:
                        playerExecutesTrigger[player] = True
    return playerExecutesTrigger


def NoWaitAndPreserved(bTrigger):
    execOnce = True
    for i in range(64):
        actionflag = bTrigger[320 + 32 * i + 28]
        if not actionflag & 0b10:  # Enabled Flag
            continue
        actionbyte = bTrigger[320 + 32 * i + 26]
        if actionbyte in (4, 7):  # Wait or Transmission
            return False
        elif actionbyte == 3:  # PreserveTrigger
            execOnce = False
    if execOnce and not (bTrigger[320 + 2048] & 4):
        return False
    return True


def InlineCodifyBinaryTrigger(bTrigger):
    """ Inline codify raw(binary) trigger data.

    For minimal protection, eudplib make some of the trig-triggers to
    eudplib trigger. This function makes eudplib trigger out of raw
    binary trigger stream.

    :param bTrigger: Binary trigger data
    :returns: (tStart, tEnd) pair, as being used by tEnd
    """

    # 1. Get executing players of the trigger.
    # If all player executes it, then pass it
    playerExecutesTrigger = GetExecutingPlayers(bTrigger)

    # 2. Create function body

    if c.PushTriggerScope():
        tStart = c.RawTrigger(actions=c.SetDeaths(0, c.SetTo, 0, 0))

        cp = _runner_cp
        execP = [p for p, e in enumerate(playerExecutesTrigger) if e]
        noWaitPreserve = NoWaitAndPreserved(bTrigger)

        if noWaitPreserve and len(execP) == 8:
            tEnd = c.RawTrigger(trigSection=bTrigger)
        if noWaitPreserve and max(execP) - min(execP) == len(execP) - 1:
            if cs.EUDIf()([cp >= min(execP), cp <= max(execP)]):
                c.RawTrigger(trigSection=bTrigger)
            cs.EUDEndIf()
            tEnd = c.RawTrigger()
        else:
            cs.EUDSwitch(cp)
            for player in ut.RandList(range(8)):
                if playerExecutesTrigger[player]:
                    if cs.EUDSwitchCase()(player):
                        c.RawTrigger(trigSection=bTrigger)
                        cs.EUDBreak()

            cs.EUDEndSwitch()

            tEnd = c.RawTrigger()
    c.PopTriggerScope()

    return (tStart, tEnd)


def CountConditionsAndActions(bTrigger):
    cond_count, act_count = 0, 0
    for c in range(16):
        if bTrigger[c * 20 + 15] == 22:  # Always
            continue
        elif bTrigger[c * 20 + 15] >= 1:
            cond_count += 1
    for a in range(64):
        if bTrigger[320 + a * 32 + 26] in (3, 47):  # PreserveTrigger, Comment
            continue
        elif bTrigger[320 + a * 32 + 26] >= 1:
            act_count += 1
    return cond_count, act_count


def GetTriggerSize(bTrigger):
    cond_count, act_count = CountConditionsAndActions(bTrigger)
    min_size = 4 + 5 * cond_count + 8 * act_count
    trig = {
        "n": [1],  # nextptr
        "P": [594],  # preserved
        "C": list(range(2, 2 + 5 * cond_count)),  # conditions
        "c": [5 + 5 * cond_count],  # condtype == 0
        "A": list(range(82, 82 + 8 * act_count)),  # actions
        "a": [88 + 8 * act_count],  # acttype == 0
    }
    if cond_count == 16:
        del trig["c"]
    if act_count == 64:
        del trig["a"]

    for ret in range(min_size, 2408 // 4 + 1):
        output = ["_" for _ in range(ret)]

        def _f():
            for k, v in trig.items():
                for x in v:
                    if output[x % ret] != "_":
                        return False
                    output[x % ret] = k
            return True

        if _f():
            return ret * 4


def TryToShareTrigger(bTrigger):
    if NoWaitAndPreserved(bTrigger):
        sharedTriggers.append(bTrigger)
        return len(sharedTriggers) - 1
    return bTrigger


def InlineCodifyMultipleBinaryTriggers(bTriggers):
    """ Inline codify raw(binary) trigger data.

    For minimal protection, eudplib make some of the trig-triggers to
    eudplib trigger. This function makes eudplib trigger out of raw
    binary trigger stream.

    :param bTrigger: Binary trigger data
    :returns: (tStart, tEnd) pair, as being used by tEnd
    """

    # Create function body

    firstTrigger = None
    tStartActions = [c.SetDeaths(0, c.SetTo, 0, 0)]
    b2s = False

    if c.PushTriggerScope():
        nextTrigger = c.Forward()
        for i, bTrigger in enumerate(bTriggers):
            if isinstance(bTrigger, bytes):
                tEnd = c.RawTrigger(trigSection=bTrigger)
                nextTrigger << tEnd
                if i < len(bTriggers) - 1:
                    nextTrigger = c.Forward()
                b2s = True
            elif isinstance(bTrigger, int):
                tEnd = sharedTriggers[bTrigger]
                if isinstance(tEnd, bytes):
                    tEnd = c.RawTrigger(trigSection=tEnd)
                    sharedTriggers[bTrigger] = tEnd
                elif isinstance(tEnd, c.RawTrigger):
                    if b2s:
                        c.SetNextTrigger(tEnd)
                else:
                    raise TypeError()
                nextTrigger << tEnd
                if i < len(bTriggers) - 1:
                    nextTrigger = c.Forward()
                    tStartActions.append(c.SetNextPtr(tEnd, nextTrigger))
                b2s = False
            else:
                raise TypeError()
            if firstTrigger is None:
                firstTrigger = tEnd
    c.PopTriggerScope()

    if c.PushTriggerScope():
        tStart = c.Forward()
        tStart << c.NextTrigger()

        cs.DoActions(tStartActions)

        c.SetNextTrigger(firstTrigger)

    c.PopTriggerScope()

    return (tStart, tEnd)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2019 Armoha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from ... import core as c, ctrlstru as cs, utils as ut
from ...trigger import Trigger
from ..memiof import (
    f_dwread_epd,
    f_wread_epd,
    f_dwbreak,
    f_getcurpl,
    f_setcurpl,
    f_memcmp,
)
from .eudprint import f_dbstr_print, ptr2s, epd2s
from .fmtprint import _format_args
from .cpprint import f_cpstr_print, PName
from .strfunc import f_strlen_epd
from ..utilf import (
    f_playerexist,
    f_getgametick,
    EUDLoopPlayer,
)
from ...localize import _
from ..playerv import PVariable
from ..eudarray import EUDArray
from math import ceil

PLVarUnit, PLVarMask = ceil((0x58F500 - 0x58A364) / 48), 0
PLVarDict = {}


def GetPlayerLightVariable():
    global PLVarUnit, PLVarMask
    ret = (PLVarUnit, 1 << PLVarMask)
    PLVarMask += 1
    if PLVarMask >= 32:
        PLVarMask = 0
        PLVarUnit += 1
    return ret


def compare_sequence(src, seq):
    assert isinstance(src, int) and isinstance(seq, str)
    seq += "\0"
    ret = []
    val, mask = 0, 0

    def append_cmp():
        nonlocal val, mask
        ret.append(c.MemoryX(src - src % 4, c.Exactly, val, mask))
        val, mask = 0, 0

    for char in seq:
        lsh = 8 * (src % 4)
        val += ord(char) << lsh
        mask += 0xFF << lsh
        if src % 4 == 3:
            append_cmp()
        src += 1
    if val and mask:
        append_cmp()
    return ret


def GetIsPNameCondition(name, _plvars={}):
    try:
        init, end, params = _plvars[name]
    except KeyError:
        c.PushTriggerScope()
        unit, mask = GetPlayerLightVariable()
        init, params = c.NextTrigger(), (c.AtLeast, 1, unit, mask)
        end = c.Forward()
        _plvars[name] = (init, end, params)
        if isinstance(name, str):
            # TODO: only iterate human players
            for player in range(8):
                t = c.RawTrigger(
                    conditions=compare_sequence(0x57EEEB + 36 * player, name),
                    actions=c.SetDeaths(player, c.Add, mask, unit),
                    preserved=False,
                )
        else:
            # TODO: compare every names in a single loop
            for player in EUDLoopPlayer(None):
                equals = f_memcmp(0x57EEEB + 36 * player, name, 25)
                Trigger(
                    conditions=equals.Exactly(0),
                    actions=c.SetDeaths(player, c.Add, mask, unit),
                    preserved=False,
                )
            t = c.RawTrigger()
        end << t
        c.PopTriggerScope()
    return init, end, params


def IsPName(player, name):
    p = c.EncodePlayer(player)
    if isinstance(p, int):
        if p < 8:
            if isinstance(name, str):
                return compare_sequence(0x57EEEB + 36 * p, name)
        else:
            ut.ep_assert(
                p == c.EncodePlayer(c.CurrentPlayer),
                _("IsPName player should be Player1 to Player8 or CurrentPlayer, not {}").format(player),
            )

    init, end, params = GetIsPNameCondition(name)
    if cs.EUDExecuteOnce()():
        # TODO: initialize when game starts
        nptr = c.Forward()
        c.RawTrigger(nextptr=init, actions=c.SetNextPtr(end, nptr))
        nptr << c.NextTrigger()
    cs.EUDEndExecuteOnce()
    return c.DeathsX(p, *params)


isPNameInitialized = False

isTxtPtrUnchanged = None
temp_Db = None
ptr, epd = None, None
odds_or_even = None
baselens = None
NameDb = None
baseNames = None
start_optimize = c.Forward()
end_optimize = c.Forward()
odd, even0, even1 = [[c.Forward() for p in range(8)] for _ in range(3)]


def _InitPName():
    global isTxtPtrUnchanged, temp_Db, ptr, epd, temp_Db, odds_or_even
    global baselens, baseNames
    isTxtPtrUnchanged = c.EUDLightVariable()
    temp_Db = c.Db(218)
    ptr, epd = c.EUDVariable(), c.EUDVariable()
    odds_or_even = c.EUDLightVariable()
    baselens = PVariable()
    NameDb = [c.Db(26) for _ in range(8)]
    baseNames = PVariable(NameDb)


@c.EUDFunc
def f_check_id(player, dst):
    ret = c.EUDVariable()
    ret << 1

    cs.EUDSwitch(player)
    for i in range(8):
        cs.EUDSwitchCase()(i)
        if cs.EUDIf()(
            [odds_or_even.Exactly(0), odd[i] << c.MemoryEPD(dst, c.Exactly, 0)]
        ):
            ret << 0

        if cs.EUDElseIf()(
            [
                odds_or_even.Exactly(1),
                even0[i] << c.MemoryXEPD(dst, c.Exactly, 0, 0xFFFF0000),
                even1[i] << c.MemoryEPD(dst + 1, c.Exactly, 0),
            ]
        ):
            ret << 0
        cs.EUDEndIf()
        cs.EUDBreak()
    cs.EUDEndSwitch()

    return ret


def _OptimizeSetPName():
    c.PushTriggerScope()
    global start_optimize, end_optimize, isTxtPtrUnchanged
    prevTxtPtr, _end = c.Forward(), c.Forward()
    start_optimize << c.NextTrigger()

    oddA = EUDArray([ut.EPD(x) for x in odd])
    even0A = EUDArray([ut.EPD(x) for x in even0])
    even1A = EUDArray([ut.EPD(x) for x in even1])

    if cs.EUDExecuteOnce()():
        for player in EUDLoopPlayer(None):
            player_name = 36 * player
            playerid_epd = 9 * player
            cs.DoActions(
                player_name.AddNumber(0x57EEEB),
                playerid_epd.AddNumber(ut.EPD(0x57EEEC)),
            )
            idlen = f_strlen_epd(playerid_epd)
            nameDb = baseNames[player]
            f_dbstr_print(nameDb, ptr2s(player_name), ":")
            val_odd = f_dwread_epd(ut.EPD(nameDb))
            v0, v1 = f_dwbreak(val_odd)[0:2]
            val_even0 = v0 * 0x10000
            val_even1 = v1 + f_wread_epd(ut.EPD(nameDb) + 1, 0) * 0x10000
            con_odd, con_even = oddA[player], even1A[player]
            cs.DoActions(
                idlen.AddNumber(1),  # fix ":" not recognized
                player_name.AddNumber(1),
                c.SetMemoryEPD(con_odd + 2, c.SetTo, val_odd),
                c.SetMemoryEPD(even0A[player] + 2, c.SetTo, val_even0),
                c.SetMemoryEPD(con_even + 2, c.SetTo, val_even1),
            )
            f_dbstr_print(nameDb, ptr2s(player_name), ":")
            baselens[player] = idlen
            if cs.EUDIf()(idlen <= 4):
                cs.DoActions(
                    c.SetMemoryEPD(con_even + 2, c.SetTo, 0),
                    c.SetMemoryEPD(con_even + 3, c.SetTo, 0x0F000000),
                )
                if cs.EUDIf()(idlen <= 2):
                    cs.DoActions(
                        c.SetMemoryEPD(con_odd + 2, c.SetTo, 0),
                        c.SetMemoryEPD(con_odd + 3, c.SetTo, 0x0F000000),
                    )
                cs.EUDEndIf()
            cs.EUDEndIf()
    cs.EUDEndExecuteOnce()

    once = c.Forward()
    cs.EUDJumpIf([once << c.Memory(0x57F23C, c.Exactly, ~0)], end_optimize)
    cs.DoActions(
        isTxtPtrUnchanged.SetNumber(0), c.SetMemory(once + 8, c.SetTo, f_getgametick()),
    )
    c.RawTrigger(
        conditions=[prevTxtPtr << c.Memory(0x640B58, c.Exactly, 11)],
        actions=isTxtPtrUnchanged.SetNumber(1),
    )
    cs.EUDJumpIf(isTxtPtrUnchanged.Exactly(1), _end)
    cs.DoActions(c.SetMemory(prevTxtPtr + 8, c.SetTo, 0))
    for i in range(3, -1, -1):
        c.RawTrigger(
            conditions=c.MemoryX(0x640B58, c.AtLeast, 1, 2 ** i),
            actions=c.SetMemory(prevTxtPtr + 8, c.Add, 2 ** i),
        )
    _end << c.NextTrigger()
    end_optimize << c.RawTrigger()
    c.PopTriggerScope()


def SetPNamef(player, format_string, *args):
    fmtargs = _format_args(format_string, *args)
    SetPName(player, *fmtargs)


def SetPName(player, *name):
    global isPNameInitialized
    if not isPNameInitialized:
        _InitPName()
        _OptimizeSetPName()
        isPNameInitialized = True

    start = c.Forward()
    c.RawTrigger(nextptr=start_optimize, actions=c.SetNextPtr(end_optimize, start))
    start << c.NextTrigger()

    _end = c.Forward()
    cs.EUDJumpIf(isTxtPtrUnchanged.Exactly(1), _end)
    if ut.isUnproxyInstance(player, type(c.P1)):
        if player == c.CurrentPlayer:
            player = f_getcurpl()
        else:
            player = c.EncodePlayer(player)
    cs.EUDJumpIf(f_playerexist(player) == 0, _end)
    basename, baselen = baseNames[player], baselens[player]

    global ptr, epd, odds_or_even
    origcp = f_getcurpl()
    cs.DoActions(
        ptr.SetNumber(0x640B61),
        epd.SetNumber(ut.EPD(0x640B60)),
        odds_or_even.SetNumber(0),
    )
    if cs.EUDWhile()(ptr.AtMost(0x640B61 + 218 * 10)):
        cs.EUDContinueIf(f_check_id(player, epd))
        cs.EUDContinueIf(f_memcmp(basename, ptr, baselen) >= 1)
        cs.DoActions(
            c.SetCurrentPlayer(ut.EPD(temp_Db)), baselen.SubtractNumber(1)
        )  # fix setpname save 1byte less chat contents
        f_cpstr_print(ptr2s(ptr + baselen), EOS=False)
        cs.DoActions(
            c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
            baselen.AddNumber(1),
            c.SetCurrentPlayer(epd),
        )
        c.RawTrigger(
            conditions=odds_or_even.Exactly(1),
            actions=[
                c.SetDeaths(c.CurrentPlayer, c.SetTo, ut.b2i4(b"\r" * 4), 0),
                c.AddCurrentPlayer(1),
            ],
        )
        name = list(name)
        name.append(epd2s(ut.EPD(temp_Db)))
        f_cpstr_print(*name)

        cs.EUDSetContinuePoint()
        cs.DoActions(
            ptr.AddNumber(218), epd.AddNumber(54), odds_or_even.AddNumberX(1, 1)
        )
        c.RawTrigger(conditions=odds_or_even.Exactly(0), actions=epd.AddNumber(1))
    cs.EUDEndWhile()
    f_setcurpl(origcp)

    _end << c.NextTrigger()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2019 Armoha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import random

from ..core.rawtrigger import (
    EncodeModifier,
    SetDeaths,
    Action,
    SetTo,
    Add,
    Subtract,
    CurrentPlayer,
)
from ..utils import EPD

_seed = 0
_stack = []


def push():
    _stack.append(_seed)


def pop():
    _stack.pop()


def load():
    global _seed
    _seed = _stack[-1]


def rand(dest):
    unit = random.randint(234, 65535)
    cpo = EPD(dest) - 12 * unit - _seed
    return cpo, unit


def srand():
    r = random.randint(27, 0xFFFFFFFF)
    _loc = random.randint(1, 0xFFFFFFFF)
    u = random.randint(234, 65535)
    epd = EPD(0x6509B0) - 12 * u
    global _seed
    _seed = r
    return Action(_loc, 0, 0, 0, epd, r, u, 45, 7, 20)


def SetMemoryS(dest, modtype, value):
    modtype = EncodeModifier(modtype, issueError=True)
    cpo, unit = rand(dest)
    return [
        SetMemoryC(0x6509B0, Add, cpo),
        SetDeaths(CurrentPlayer, modtype, value, unit),
    ]


def MoveCP(dest):
    try:
        value = dest - _seed
    except TypeError:
        value = _seed * (-1) + dest
    return SetMemoryC(0x6509B0, Add, value)


def SetMemoryC(dest, modtype, value):
    modtype = EncodeModifier(modtype, issueError=True)
    _loc = random.randint(0, 0xFFFFFFFF)
    u = random.randint(234, 65535)
    epd = EPD(dest) - 12 * u
    if dest == 0x6509B0:
        global _seed
        if modtype == 7:
            _seed = value
        elif modtype == 9:
            try:
                _seed = _seed - value
            except TypeError:
                _seed = -value + _seed
        else:
            try:
                _seed = _seed + value
            except TypeError:
                _seed = value + _seed
    flag = random.randint(0, 0xFF) & (0xFF - 2)
    return Action(_loc, 0, 0, 0, epd, value, u, 45, modtype, flag)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2019 Armoha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...core.mapdata.stringmap import (
    ForceAddString,
    ApplyStringMap,
    GetStringMap,
)
from ..eudarray import EUDArray
from ..memiof import f_dwread_epd, f_getcurpl, f_setcurpl, f_wread_epd
from ..utilf import f_getuserplayerid, IsUserCP
from .cpprint import f_cpstr_print, f_gettextptr
from .cpstr import GetMapStringAddr
from .fmtprint import _format_args
from .strfunc import f_strlen_epd
from .texteffect import TextFX_FadeIn, TextFX_FadeOut, TextFX_Remove


@c.EUDTypedFunc([None, c.TrgString])
def DisplayTextAt(line, text):
    tp = f_gettextptr()
    c.VProc(line, line.QueueAddTo(ut.EPD(0x640B58)))
    c.RawTrigger(
        conditions=c.Memory(0x640B58, c.AtLeast, 11),
        actions=c.SetMemory(0x640B58, c.Subtract, 11),
    )
    cs.DoActions(c.DisplayText(text), c.SetMemory(0x640B58, c.SetTo, tp))


class StringBuffer:
    """Object for storing single modifiable string.

    Manipluating STR section is easy. :)
    You can do anything you would do with normal string with StringBuffer.
    """

    _method_template = c.Forward()
    _cpbranch = c.Forward()

    def __init__(self, content=None):
        """Constructor for StringBuffer

        :param content: Initial StringBuffer content / capacity. Capacity of
            StringBuffer is determined by size of this. If content is integer, then
            initial capacity and content of StringBuffer will be set to
            content(int) and empty string.

        :type content: str, bytes, int
        """
        chkt = c.GetChkTokenized()
        if content is None:
            content = "\r" * 218
        elif isinstance(content, int):
            content = "\r" * content
        else:
            content = ut.u2utf8(content)
        self.capacity = len(content)
        self.StringIndex = ForceAddString(content)
        self.epd, self.pos = c.EUDVariable(), c.EUDVariable()

        try:
            cs.EUDExecuteOnce()()
            cs.DoActions(self.epd.SetNumber(ut.EPD(GetMapStringAddr(self.StringIndex))))
            cs.EUDEndExecuteOnce()
        except IndexError:
            from ...maprw.injector.mainloop import EUDOnStart

            def _f():
                cs.DoActions(
                    self.epd.SetNumber(ut.EPD(GetMapStringAddr(self.StringIndex)))
                )

            EUDOnStart(_f)

    @classmethod
    def _init_template(cls):
        c.PushTriggerScope()
        cls._method_template << c.NextTrigger()
        localcp = f_getuserplayerid()
        cls._cpbranch << c.RawTrigger(
            nextptr=0, conditions=IsUserCP(), actions=c.SetNextPtr(cls._cpbranch, 0),
        )
        c.PopTriggerScope()

    @classmethod
    def CPBranch(cls):
        if not StringBuffer._method_template.IsSet():
            StringBuffer._init_template()
        end, ontrue = c.Forward(), c.Forward()
        c.RawTrigger(
            nextptr=StringBuffer._method_template,
            actions=[
                c.SetNextPtr(StringBuffer._cpbranch, end),
                c.SetMemory(StringBuffer._cpbranch + 348, c.SetTo, ontrue),
            ],
        )
        return end, ontrue

    def append(self, *args):
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        f_setcurpl(self.pos)
        f_cpstr_print(*args)
        f_getcurpl(ret=[self.pos])
        f_setcurpl(f_getuserplayerid())
        end << c.NextTrigger()

    def appendf(self, format_string, *args):
        fmtargs = _format_args(format_string, *args)
        self.append(*fmtargs)

    def insert(self, index, *args):
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        f_setcurpl(self.epd + index)
        f_cpstr_print(*args, EOS=False)
        f_getcurpl(ret=[self.pos])
        f_setcurpl(f_getuserplayerid())
        end << c.NextTrigger()

    def insertf(self, index, format_string, *args):
        fmtargs = _format_args(format_string, *args)
        self.insert(index, *fmtargs)

    def delete(self, start, length=1):
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        self.pos << self.epd
        self.pos += start
        f_setcurpl(self.pos)
        cs.DoActions(
            [
                c.SetDeaths(c.CurrentPlayer, c.SetTo, 0x0D0D0D0D, 0),
                c.AddCurrentPlayer(1),
            ]
            for _ in range(length)
        )
        f_setcurpl(f_getuserplayerid())
        end << c.NextTrigger()

    def Display(self):
        cs.DoActions(c.DisplayText(self.StringIndex))

    def DisplayAt(self, line):
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        prevptr = f_gettextptr()
        cs.DoActions(c.SetMemory(0x640B58, c.Add, line))
        c.RawTrigger(
            conditions=c.Memory(0x640B58, c.AtLeast, 11),
            actions=c.SetMemory(0x640B58, c.Subtract, 11),
        )
        c.VProc(
            prevptr,
            [c.DisplayText(self.StringIndex), prevptr.SetDest(ut.EPD(0x640B58))],
        )
        end << c.NextTrigger()

    def print(self, *args):
        if len(args) == 1 and isinstance(args[0], str) and len(args[0]) > 31:
            cs.DoActions(c.DisplayText(args[0]))
            return
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        f_setcurpl(self.epd)
        f_cpstr_print(*args, EOS=False)
        cs.DoActions(
            c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
            c.SetCurrentPlayer(f_getuserplayerid()),
            c.DisplayText(self.StringIndex),
        )
        end << c.NextTrigger()

    def printf(self, format_string, *args):
        fmtargs = _format_args(format_string, *args)
        self.print(*fmtargs)

    def printAt(self, line, *args):
        end, ontrue = StringBuffer.CPBranch()
        ontrue << c.NextTrigger()
        f_setcurpl(self.epd)
        f_cpstr_print(*args, EOS=False)
        cs.DoActions(
            c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
            c.SetCurrentPlayer(f_getuserplayerid()),
        )
        self.DisplayAt(line)
        end << c.NextTrigger()

    def printfAt(self, line, format_string, *args):
        fmtargs = _format_args(format_string, *args)
        self.printAt(line, *fmtargs)

    def Play(self):
        cs.DoActions(c.PlayWAV(self.StringIndex))

    def fadeIn(self, *args, color=None, wait=1, reset=True, line=-1, tag=None):
        if not StringBuffer._method_template.IsSet():
            StringBuffer._init_template()
        end, ontrue = c.Forward(), c.Forward()
        ret = c.EUDVariable()
        c.RawTrigger(
            nextptr=StringBuffer._method_template,
            actions=[
                c.SetNextPtr(StringBuffer._cpbranch, end),
                c.SetMemory(StringBuffer._cpbranch + 348, c.SetTo, ontrue),
                ret.SetNumber(-1),
            ],
        )
        ontrue << c.NextTrigger()
        prevpos = TextFX_Remove(tag)
        f_setcurpl(self.epd)
        ret << TextFX_FadeIn(*args, color=color, wait=wait, reset=reset, tag=tag)
        f_setcurpl(f_getuserplayerid())
        if type(line) == int:
            if 0 <= line <= 10:
                self.DisplayAt(line)
            elif line >= -11:
                if cs.EUDIf()(prevpos == -1):
                    self.DisplayAt(11 + line)
                if cs.EUDElse()():
                    prevptr = f_gettextptr()
                    c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                    c.VProc(
                        prevptr,
                        [
                            c.DisplayText(self.StringIndex),
                            prevptr.SetDest(ut.EPD(0x640B58)),
                        ],
                    )
                cs.EUDEndIf()
        else:
            if cs.EUDIf()(line <= 10):
                self.DisplayAt(line)
            if cs.EUDElseIf()(line >= 0xFFFFFFF5):
                if cs.EUDIf()(prevpos == -1):
                    self.DisplayAt(11 + line)
                if cs.EUDElse()():
                    prevptr = f_gettextptr()
                    c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                    c.VProc(
                        prevptr,
                        [
                            c.DisplayText(self.StringIndex),
                            prevptr.SetDest(ut.EPD(0x640B58)),
                        ],
                    )
                cs.EUDEndIf()
            cs.EUDEndIf()
        end << c.NextTrigger()
        return ret

    def fadeInf(
        self, format_string, *args, color=None, wait=1, reset=True, line=-1, tag=None
    ):
        fmtargs = _format_args(format_string, *args)
        return self.fadeIn(
            *fmtargs, color=color, wait=wait, reset=reset, line=line, tag=tag
        )

    def fadeOut(self, *args, color=None, wait=1, reset=True, line=-1, tag=None):
        if not StringBuffer._method_template.IsSet():
            StringBuffer._init_template()
        end, ontrue = c.Forward(), c.Forward()
        ret = c.EUDVariable()
        c.RawTrigger(
            nextptr=StringBuffer._method_template,
            actions=[
                c.SetNextPtr(StringBuffer._cpbranch, end),
                c.SetMemory(StringBuffer._cpbranch + 348, c.SetTo, ontrue),
                ret.SetNumber(-1),
            ],
        )
        ontrue << c.NextTrigger()
        prevpos = TextFX_Remove(tag)
        f_setcurpl(self.epd)
        ret << TextFX_FadeOut(*args, color=color, wait=wait, reset=reset, tag=tag)
        f_setcurpl(f_getuserplayerid())
        if type(line) == int:
            if 0 <= line <= 10:
                self.DisplayAt(line)
            elif line >= -11:
                if cs.EUDIf()(prevpos == -1):
                    self.DisplayAt(11 + line)
                if cs.EUDElse()():
                    prevptr = f_gettextptr()
                    c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                    c.VProc(
                        prevptr,
                        [
                            c.DisplayText(self.StringIndex),
                            prevptr.SetDest(ut.EPD(0x640B58)),
                        ],
                    )
                cs.EUDEndIf()
        else:
            if cs.EUDIf()(line <= 10):
                self.DisplayAt(line)
            if cs.EUDElseIf()(line >= 0xFFFFFFF5):
                if cs.EUDIf()(prevpos == -1):
                    self.DisplayAt(11 + line)
                if cs.EUDElse()():
                    prevptr = f_gettextptr()
                    c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                    c.VProc(
                        prevptr,
                        [
                            c.DisplayText(self.StringIndex),
                            prevptr.SetDest(ut.EPD(0x640B58)),
                        ],
                    )
                cs.EUDEndIf()
            cs.EUDEndIf()
        end << c.NextTrigger()
        return ret

    def fadeOutf(
        self, format_string, *args, color=None, wait=1, reset=True, line=-1, tag=None
    ):
        fmtargs = _format_args(format_string, *args)
        return self.fadeOut(
            *fmtargs, color=color, wait=wait, reset=reset, line=line, tag=tag
        )

    def length(self):
        return f_strlen_epd(self.epd)


_globalsb = None


def GetGlobalStringBuffer():
    global _globalsb
    if _globalsb is None:
        _globalsb = StringBuffer(1023)
    return _globalsb


def f_simpleprint(*args, spaced=True):
    # Add spaces between arguments
    if spaced:
        spaced_args = []
        for arg in args:
            spaced_args.extend([arg, " "])
        args = spaced_args[:-1]

    # Print
    oldcp = f_getcurpl()
    f_setcurpl(f_getuserplayerid())
    gsb = GetGlobalStringBuffer()
    gsb.print(*args)
    f_setcurpl(oldcp)


def f_println(format_string, *args):
    gsb = GetGlobalStringBuffer()
    gsb.printf(format_string, *args)


def f_printAt(line, format_string, *args):
    gsb = GetGlobalStringBuffer()
    gsb.printfAt(line, format_string, *args)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2019 Armoha

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import itertools
from ... import core as c, ctrlstru as cs, utils as ut
from ..memiof import f_getcurpl, f_setcurpl, f_setcurpl2cpcache
from .eudprint import ptr2s, epd2s
from .cpprint import f_cpstr_print, cw
from ..stringf.rwcommon import br1
from ..utilf import f_getgametick


color_codes = list(range(1, 32))
color_codes.remove(0x9)  # tab
color_codes.remove(0xA)  # linebreak
color_codes.remove(0xC)  # linebreak
color_codes.remove(0x12)  # right align
color_codes.remove(0x13)  # center align
color_v = c.EUDVariable(2)
color_code = b"\x02"


@c.EUDFunc
def _cpchar_addstr():
    if cs.EUDInfLoop()():
        b1 = br1.readbyte()
        cs.EUDBreakIf(b1 == 0)
        cw.writebyte(color_v)
        cw.writebyte(b1)
        if cs.EUDIf()(b1 <= 0x7F):
            cw.flushdword()
        if cs.EUDElse()():
            b2 = br1.readbyte()
            cw.writebyte(b2)
            if cs.EUDIf()(b1 <= 0xDF):  # Encode as 2-byte
                cw.flushdword()
            if cs.EUDElse()():  # 3-byte
                cw.writebyte(br1.readbyte())
            cs.EUDEndIf()
        cs.EUDEndIf()
    cs.EUDEndInfLoop()


def f_cpchar_addstr(src):
    br1.seekoffset(src)
    _cpchar_addstr()


def f_cpchar_addstr_epd(src):
    br1.seekepd(src)
    _cpchar_addstr()


@c.EUDFunc
def f_cpchar_adddw(number):
    skipper = [c.Forward() for _ in range(9)]
    ch = [0] * 10

    for i in range(10):  # Get digits
        number, ch[i] = c.f_div(number, 10)
        if i != 9:
            cs.EUDJumpIf(number == 0, skipper[i])

    for i in range(9, -1, -1):  # print digits
        if i != 9:
            skipper[i] << c.NextTrigger()
        cs.DoActions(
            c.SetDeaths(
                c.CurrentPlayer, c.SetTo, color_v + ch[i] * 256 + (0x0D0D3000), 0
            ),
            c.AddCurrentPlayer(1),
        )


def f_cpchar_print(*args, EOS=True, encoding="UTF-8"):
    global color_code
    args = ut.FlattenList(args)

    if encoding.upper() == "UTF-8":
        encode_func = ut.u2utf8
    else:
        encode_func = ut.u2b

    for arg in args:
        if isinstance(arg, bytes):
            arg = arg.decode(encoding)
        if isinstance(arg, str):
            bytestring = b""
            new_color_code = color_code
            for char in arg:
                char = encode_func(char)
                if ut.b2i1(char) in color_codes:
                    new_color_code = char
                    continue
                while len(char) < 3:
                    char = char + b"\r"
                bytestring = bytestring + new_color_code + char
            if not bytestring:
                bytestring = color_code + b"\r\r\r"
            f_cpstr_print(bytestring, EOS=False)
            if color_code != new_color_code:
                color_code = new_color_code
                cs.DoActions(color_v.SetNumber(ut.b2i1(color_code)))
        elif ut.isUnproxyInstance(arg, ptr2s):
            f_cpchar_addstr(arg._value)
        elif ut.isUnproxyInstance(arg, epd2s):
            f_cpchar_addstr_epd(arg._value)
        elif ut.isUnproxyInstance(arg, c.EUDVariable) or c.IsConstExpr(arg):
            f_cpchar_adddw(arg)
        else:
            f_cpstr_print(arg, EOS=False, encoding=encoding)
    if EOS:
        cs.DoActions(c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0))


_TextFX_dict = dict()
id_codes = "\x18\x02\x03\x04\x06\x07\x08\x0D\x0E\x0F\x10\x11\x15\x16\x17\x19\x1A\x1B\x1C\x1D\x1E\x1F"
id_gen = itertools.cycle(itertools.product(id_codes, repeat=6))


def _add_TextFX_timer(tag):
    if tag not in _TextFX_dict:
        _TextFX_dict[tag] = (
            c.EUDVariable(),
            c.EUDLightVariable(),
            ut.u2b("".join(next(id_gen))),
        )


def TextFX_SetTimer(tag, modtype, value):
    _add_TextFX_timer(tag)
    timer, _, _ = _TextFX_dict[tag]
    cs.DoActions(c.SetMemory(timer.getValueAddr(), modtype, value))


@c.EUDFunc
def _remove_TextFX(o0, o1, e0, e1):
    txtPtr = c.EUDVariable()
    trg_o0, trg_o1, trg_e0, trg_e1 = [c.Forward() for _ in range(4)]
    setPtr_o, setPtr_e = c.Forward(), c.Forward()

    c.VProc(
        [o0, o1, e0, e1],
        [
            txtPtr.SetNumber(-1),
            c.SetMemory(0x6509B0, c.SetTo, ut.EPD(0x640B60)),
            c.SetMemory(setPtr_o + 20, c.SetTo, 0),
            c.SetMemory(setPtr_e + 20, c.SetTo, 1),
            o0.SetDest(ut.EPD(trg_o0 + 16)),
            o1.SetDest(ut.EPD(trg_o1 + 16)),
            e0.SetDest(ut.EPD(trg_e0 + 16)),
            e1.SetDest(ut.EPD(trg_e1 + 16)),
        ],
    )
    if cs.EUDLoopN()(6):
        incr_o, incr_e = c.Forward(), c.Forward()
        trg_o0 << c.RawTrigger(
            nextptr=incr_o,
            conditions=[c.Deaths(c.CurrentPlayer, c.Exactly, 0, 0)],
            actions=[c.SetMemory(0x6509B0, c.Add, 1), c.SetNextPtr(trg_o0, trg_o1)],
        )
        trg_o1 << c.RawTrigger(
            conditions=[c.DeathsX(c.CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)],
            actions=[
                c.SetNextPtr(trg_o0, incr_o),
                c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
                c.SetMemory(0x6509B0, c.Add, -1),
                c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
                setPtr_o << txtPtr.SetNumber(0),
            ],
        )
        incr_o << c.RawTrigger(
            actions=[
                c.SetMemory(0x6509B0, c.Add, 216 // 4),
                c.SetMemory(setPtr_o + 20, c.Add, 2),
            ]
        )
        trg_e0 << c.RawTrigger(
            nextptr=incr_e,
            conditions=[c.DeathsX(c.CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000)],
            actions=[c.SetMemory(0x6509B0, c.Add, 1), c.SetNextPtr(trg_e0, trg_e1)],
        )
        trg_e1 << c.RawTrigger(
            conditions=[c.Deaths(c.CurrentPlayer, c.Exactly, 0, 0)],
            actions=[
                c.SetNextPtr(trg_e0, incr_e),
                c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0),
                c.SetMemory(0x6509B0, c.Add, -1),
                c.SetDeathsX(c.CurrentPlayer, c.SetTo, 0, 0, 0xFFFF0000),
                setPtr_e << txtPtr.SetNumber(1),
            ],
        )
        incr_e << c.RawTrigger(
            actions=[
                c.SetMemory(0x6509B0, c.Add, 220 // 4),
                c.SetMemory(setPtr_e + 20, c.Add, 2),
            ]
        )
    cs.EUDEndLoopN()
    f_setcurpl2cpcache()
    return txtPtr


def TextFX_Remove(tag):
    _add_TextFX_timer(tag)
    _, _, identifier = _TextFX_dict[tag]
    o0 = ut.b2i4(identifier[0:4])
    o1 = ut.b2i2(identifier[4:6])
    e0 = ut.b2i2(identifier[0:2]) * 0x10000
    e1 = ut.b2i4(identifier[2:6])
    return _remove_TextFX(o0, o1, e0, e1)


_check_cp = c.Forward()
_is_below_start = c.Forward()
_cpbelowbuffer = c.EUDLightVariable()


def _is_CP_less_than_start(actions):
    global _is_below_start
    if not _is_below_start.IsSet():
        c.PushTriggerScope()
        _is_below_start << c.RawTrigger(
            conditions=[_check_cp << c.Memory(0x6509B0, c.AtMost, 1)],
            actions=_cpbelowbuffer.SetNumber(1),
        )
        c.PopTriggerScope()
    _next = c.Forward()
    c.RawTrigger(
        nextptr=_is_below_start,
        actions=[actions]
        + [_cpbelowbuffer.SetNumber(0), c.SetNextPtr(_is_below_start, _next)],
    )
    _next << c.NextTrigger()


def R2L(colors, colors_dict={}):
    try:
        _f = colors_dict[colors]
    except (KeyError):

        @c.EUDFunc
        def _f():
            _jump, _isend, _end = [c.Forward() for _ in range(3)]
            ret = c.EUDVariable()
            _is_CP_less_than_start([ret.SetNumber(1), c.SetNextPtr(_isend, _jump)])
            _isend << c.RawTrigger(
                conditions=_cpbelowbuffer.Exactly(1),
                actions=[ret.SetNumber(0), c.SetNextPtr(_isend, _end)],
            )
            _jump << c.NextTrigger()
            for color in reversed(colors):
                _is_CP_less_than_start([])
                c.RawTrigger(
                    conditions=_cpbelowbuffer.Exactly(0),
                    actions=[
                        c.SetDeathsX(c.CurrentPlayer, c.SetTo, color, 0, 0xFF),
                        c.AddCurrentPlayer(-1),
                    ],
                )
            _end << c.NextTrigger()
            return ret

        colors_dict[colors] = _f
    return _f()


def _TextFX_Print(*args, identifier=None, encoding="UTF-8"):
    f_cpstr_print(identifier, EOS=False, encoding=encoding)
    global color_code
    color_code = b"\x02"
    color_v << 2

    args = ut.FlattenList(args)
    for arg in args:
        if isinstance(arg, str):
            line = arg.split("\n")
            if len(line) == 1:
                f_cpchar_print(line[0], EOS=False, encoding=encoding)
                continue
            for s in line[:-1]:
                f_cpchar_print(s, EOS=False, encoding=encoding)
                f_cpstr_print(
                    identifier + b"\n" + identifier, EOS=False, encoding=encoding
                )
            f_cpchar_print(line[-1], EOS=False, encoding=encoding)
        else:
            f_cpchar_print(arg, EOS=False, encoding=encoding)

    cs.DoActions(c.SetDeaths(c.CurrentPlayer, c.SetTo, 0, 0))


def TextFX_FadeIn(*args, color=None, wait=1, reset=True, tag=None, encoding="UTF-8"):
    """Print multiple string / number and apply color from Left To Right

    Keyword arguments:
    color -- tuple of color codes (default 0x03, 0x04, 0x05, 0x14)
    wait  -- time interval between effect (default 1)
    reset -- automatically reset when didn't run for a moment (default True)
    tag   -- internal tag and identifier (default: vargs)
    """
    if color is None:
        color = (0x03, 0x04, 0x05, 0x14)
    if tag is None:
        if len(args) == 1:
            tag = args[0]
        else:
            tag = args

    _add_TextFX_timer(tag)
    timer, counter, identifier = _TextFX_dict[tag]

    start = f_getcurpl()
    c.SeqCompute(
        [(ut.EPD(_check_cp) + 2, c.SetTo, 1), (ut.EPD(_check_cp) + 2, c.Add, start)]
    )
    _TextFX_Print(*args, identifier=identifier, encoding=encoding)
    f_setcurpl(start + (3 - len(color)))

    if reset is True:
        check_gametick = c.Forward()
        if cs.EUDIf()([check_gametick << c.Memory(0x57F23C, c.AtLeast, 0)]):
            gametick = f_getgametick()
            c.SeqCompute(
                [
                    (timer, c.SetTo, 0),
                    (ut.EPD(check_gametick) + 2, c.SetTo, 1),
                    (ut.EPD(check_gametick) + 2, c.Add, gametick),
                ]
            )
        cs.EUDEndIf()

    _end = c.Forward()
    _is_finished, _draw_color = c.Forward(), c.Forward()
    ret = c.EUDVariable()

    cs.DoActions(
        counter.AddNumber(1),
        c.SetMemory(check_gametick + 8, c.Add, 1) if reset is True else [],
        ret.SetNumber(1),
        c.SetNextPtr(_is_finished, _draw_color),
        c.AddCurrentPlayer(timer),
    )
    _is_finished << c.RawTrigger(
        conditions=[
            c.Deaths(c.CurrentPlayer, c.Exactly, 0, 0),
            timer.AtLeast(2 + len(color)),
        ],
        actions=[
            ret.SetNumber(0),
            counter.SetNumber(0),
            c.SetNextPtr(_is_finished, _end),
        ],
    )
    _draw_color << c.RawTrigger(actions=c.AddCurrentPlayer(len(color) - 1))
    R2L(color)
    c.RawTrigger(
        conditions=counter.AtLeast(max(wait, 1)),
        actions=[counter.SetNumber(0), timer.AddNumber(1)],
    )
    _end << c.NextTrigger()
    return ret


def TextFX_FadeOut(*args, color=None, wait=1, reset=True, tag=None, encoding="UTF-8"):
    """Print multiple string / number and apply color from Left To Right

    Keyword arguments:
    color -- tuple of color codes (default 0x03, 0x04, 0x05, 0x14)
    wait  -- time interval between effect (default 1)
    reset -- automatically reset when didn't run for a moment (default True)
    tag   -- internal tag and identifier (default: vargs)
    """
    if color is None:
        color = (0x03, 0x04, 0x05, 0x14)
    if tag is None:
        if len(args) == 1:
            tag = args[0]
        else:
            tag = args

    _add_TextFX_timer(tag)
    timer, counter, identifier = _TextFX_dict[tag]

    start = f_getcurpl()
    c.SeqCompute(
        [(ut.EPD(_check_cp) + 2, c.SetTo, 1), (ut.EPD(_check_cp) + 2, c.Add, start)]
    )
    _TextFX_Print(*args, identifier=identifier, encoding=encoding)

    if reset is True:
        check_gametick = c.Forward()
        if cs.EUDIf()([check_gametick << c.Memory(0x57F23C, c.AtLeast, 0)]):
            gametick = f_getgametick()
            c.SeqCompute(
                [
                    (timer, c.SetTo, 0),
                    (ut.EPD(check_gametick) + 2, c.SetTo, 1),
                    (ut.EPD(check_gametick) + 2, c.Add, gametick),
                ]
            )
        cs.EUDEndIf()

    cs.DoActions(
        counter.AddNumber(1),
        c.SetMemory(check_gametick + 8, c.Add, 1) if reset is True else [],
        c.AddCurrentPlayer((len(color) - 1) - timer),
    )
    ret = R2L(color)
    c.RawTrigger(conditions=ret.Exactly(0), actions=counter.SetNumber(0))
    c.RawTrigger(
        conditions=counter.AtLeast(max(wait, 1)),
        actions=[counter.SetNumber(0), timer.AddNumber(1)],
    )
    return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import functools

from ... import utils as ut

from .. import allocator as ac
from .. import rawtrigger as bt
from .. import variable as ev
from .trace.tracetool import _EUDTracePush, _EUDTracePop
from ...localize import _
from ...utils.blockstru import BlockStruManager, SetCurrentBlockStruManager


_currentCompiledFunc = None
_currentTriggerCount = 0


def _updateFuncTriggerCount():
    global _currentTriggerCount
    currentCounter = bt.GetTriggerCounter()
    addedTriggerCount = currentCounter - _currentTriggerCount

    if _currentCompiledFunc:
        _currentCompiledFunc._triggerCount += addedTriggerCount
    _currentTriggerCount = currentCounter


def _setCurrentCompiledFunc(func):
    global _currentCompiledFunc

    lastCompiledFunc = _currentCompiledFunc
    _updateFuncTriggerCount()
    _currentCompiledFunc = func
    return lastCompiledFunc


class EUDFuncN:
    def __init__(self, argn, callerfunc, bodyfunc, *, traced):
        """ EUDFuncN

        :param callerfunc: Function to be wrapped.
        :param argn: Count of arguments got by callerfunc
        :param bodyfunc: Where function should return to
        """

        if bodyfunc is None:
            bodyfunc = callerfunc

        self._argn = argn
        self._retn = None
        self._callerfunc = callerfunc
        self._bodyfunc = bodyfunc
        functools.update_wrapper(self, bodyfunc)
        self._fstart = None
        self._fend = None
        self._nptr = None
        self._fargs = None
        self._frets = None
        self._triggerCount = None
        self._traced = traced

    def size(self):
        if not self._fstart:
            self._CreateFuncBody()

        return self._triggerCount

    def _CreateFuncBody(self):
        self._triggerCount = 0
        lastCompiledFunc = _setCurrentCompiledFunc(self)

        # Add return point
        self._fend = ac.Forward()

        # Prevent double compilication
        ut.ep_assert(self._fstart is None)

        # Initalize new namespace
        f_bsm = BlockStruManager()
        prev_bsm = SetCurrentBlockStruManager(f_bsm)
        bt.PushTriggerScope()

        self._CreateFuncArgs()

        fstart = bt.NextTrigger()
        self._fstart = fstart

        if self._traced:
            _EUDTracePush()

        final_rets = self._callerfunc(*self._fargs)
        if final_rets is not None:
            self._AddReturn(ut.Assignable2List(final_rets), False)

        self._fend << bt.NextTrigger()
        if self._traced:
            _EUDTracePop()
        if self._retn is None or self._retn == 0:
            self._fend = bt.RawTrigger()
            self._nptr = self._fend + 4
        else:
            fendTrgs = ut.FlattenList(ev.VProc([self._frets], []))
            self._fend, self._nptr = fendTrgs[0], fendTrgs[-1]._actions[-1] + 20

        bt.PopTriggerScope()

        # Finalize
        ut.ep_assert(f_bsm.empty(), _("Block start/end mismatch inside function"))
        SetCurrentBlockStruManager(prev_bsm)

        # No return -> set return count to 0
        if self._retn is None:
            self._retn = 0
        _setCurrentCompiledFunc(lastCompiledFunc)

    def _CreateFuncArgs(self):
        if self._fargs is None:
            self._fargs = [ev.EUDVariable() for _ in range(self._argn)]

    def _AddReturn(self, retv, needjump):
        retv = ut.FlattenList(retv)
        if self._frets is None:
            self._frets = [ev.EUDVariable() for _ in retv]
            self._retn = len(retv)

        ut.ep_assert(
            len(retv) == len(self._frets),
            _("Number of returned values should be constant.")
            + _(" (From function %s)").format(self._bodyfunc.__name__),
        )

        constAssigns = [
            (fret, bt.SetTo, ret)
            for fret, ret in zip(self._frets, retv)
            if not ev.IsEUDVariable(ret)
        ]
        varAssigns = [
            (fret, bt.SetTo, ret)
            for fret, ret in zip(self._frets, retv)
            if ev.IsEUDVariable(ret)
        ]
        if len(varAssigns) <= 2:
            ev.SeqCompute(constAssigns + varAssigns)
        else:
            ev.NonSeqCompute(constAssigns + varAssigns)

        if needjump:
            bt.SetNextTrigger(self._fend)

    def __call__(self, *args, ret=None):
        if self._fstart is None:
            self._CreateFuncBody()

        ut.ep_assert(
            len(args) == self._argn,
            _("Argument number mismatch : ") + "len(%s) != %d" % (repr(args), self._argn),
        )

        fcallend = ac.Forward()

        # SeqCompute gets faster when const-assignments are in front of
        # variable assignments.
        if ret is None:
            ret = [ev.EUDVariable() for _ in range(self._retn)]
        ret = ut.FlattenList(ret)
        nextPtrAssignment = [(ut.EPD(self._nptr), bt.SetTo, fcallend)]
        if self._retn >= 1:
            for fret, retv in zip(self._frets, ret):
                try:
                    retv = ut.EPD(retv.getValueAddr())
                except AttributeError:
                    pass
                nextPtrAssignment.append((ut.EPD(fret.getDestAddr()), bt.SetTo, retv))
        constAssigns = [
            (farg, bt.SetTo, arg)
            for farg, arg in zip(self._fargs, args)
            if not ev.IsEUDVariable(arg)
        ]
        varAssigns = [
            (farg, bt.SetTo, arg)
            for farg, arg in zip(self._fargs, args)
            if ev.IsEUDVariable(arg)
        ]
        if len(varAssigns) <= 2:
            ev.SeqCompute(nextPtrAssignment + constAssigns + varAssigns)
        else:
            ev.NonSeqCompute(nextPtrAssignment + constAssigns + varAssigns)
        bt.SetNextTrigger(self._fstart)

        fcallend << bt.NextTrigger()

        if self._frets is not None:
            for retv in ret:
                try:
                    retv.makeR()
                except AttributeError:
                    pass
            return ut.List2Assignable(ret)


def EUDReturn(*args):
    _currentCompiledFunc._AddReturn(args, True)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .. import rawtrigger as bt
from ..allocator import Forward, ConstExpr, IsConstExpr

from ...localize import _
from ...utils import EPD, ExprProxy, ep_assert, cachedfunc, isUnproxyInstance, ep_assert

from ..variable import EUDVariable, SeqCompute, VProc
from ..variable.vbuf import GetCurrentVariableBuffer


@cachedfunc
def EUDVArrayData(size):
    ep_assert(isinstance(size, int))

    class _EUDVArrayData(ConstExpr):
        def __init__(self, initvars):
            super().__init__(self)
            ep_assert(
                len(initvars) == size,
                _("{} items expected, got {}").format(size, len(initvars)),
            )
            self._initvars = initvars

        def Evaluate(self):
            evb = GetCurrentVariableBuffer()
            if self not in evb._vdict:
                evb.CreateMultipleVarTriggers(self, self._initvars)

            return evb._vdict[self].Evaluate()

    return _EUDVArrayData


_index = EUDVariable()
_goto_getidx = [Forward() for _ in range(32)]
_ret_getidx, _end_getidx = Forward(), Forward()


@cachedfunc
def EUDVArray(size, basetype=None):
    ep_assert(isinstance(size, int))

    class _EUDVArray(ExprProxy):
        def __init__(self, initvars=None, *, _from=None):
            # Initialization from value
            if _from is not None:
                if IsConstExpr(_from):
                    baseobj = _from

                # Initialization by variable reference
                else:
                    baseobj = EUDVariable()
                    baseobj << _from

            else:
                # Int -> interpret as sequence of 0s
                if initvars is None:
                    initvars = [0] * size

                # For python iterables
                baseobj = EUDVArrayData(size)(initvars)

            super().__init__(baseobj)
            self.dontFlatten = True
            self._epd = EPD(self)
            self._basetype = basetype

        def get(self, i):
            if isUnproxyInstance(i, EUDVariable):
                r = self._eudget(i)
            else:
                ep_assert(i < size, _("EUDVArray index out of bounds."))
                if isUnproxyInstance(self, EUDVariable):
                    r = self._get(i)
                else:
                    r = self._constget(i)

            if self._basetype:
                r = self._basetype.cast(r)
            return r

        def _eudget(self, i):
            global _goto_getidx, _ret_getidx, _end_getidx
            if not _end_getidx.IsSet():
                bt.PushTriggerScope()

                for t in range(31, -1, -1):
                    _goto_getidx[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2 ** t),
                        actions=[
                            bt.SetMemory(_end_getidx + 4, bt.Add, 72 * (2 ** t)),
                            bt.SetMemory(_ret_getidx + 16, bt.Add, 18 * (2 ** t)),
                            bt.SetMemory(_ret_getidx + 48, bt.Add, 18 * (2 ** t)),
                        ],
                    )
                _end_getidx << bt.RawTrigger(
                    nextptr=0,
                    actions=[
                        _ret_getidx << bt.SetDeaths(0, bt.SetTo, 0, 0),
                        bt.SetDeaths(0, bt.SetTo, 0, 0),
                    ],
                )
                bt.PopTriggerScope()

            r = EUDVariable()
            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()

            SeqCompute(
                [
                    (EPD(_end_getidx + 4), bt.SetTo, self),
                    (EPD(_ret_getidx + 16), bt.SetTo, self._epd + 344 // 4),
                    (EPD(_ret_getidx + 48), bt.SetTo, self._epd + 1),
                ]
            )
            bt.RawTrigger(
                nextptr=i.GetVTable(),
                actions=[
                    i.QueueAssignTo(_index),
                    bt.SetNextPtr(i.GetVTable(), _goto_getidx[bits]),
                    bt.SetMemory(_ret_getidx + 20, bt.SetTo, EPD(r.getValueAddr())),
                    bt.SetMemory(_ret_getidx + 52, bt.SetTo, nptr),
                ],
            )

            nptr << bt.NextTrigger()
            return r

        def _get(self, i):
            # This function is hand-optimized

            r = EUDVariable()
            vtproc = Forward()
            nptr = Forward()
            a0, a2 = Forward(), Forward()

            SeqCompute(
                [
                    (EPD(vtproc + 4), bt.SetTo, self + 72 * i),
                    (EPD(a0 + 16), bt.SetTo, self._epd + (18 * i + 344 // 4)),
                    (EPD(a2 + 16), bt.SetTo, self._epd + (18 * i + 1)),
                ]
            )
            vtproc << bt.RawTrigger(
                nextptr=0,
                actions=[
                    a0 << bt.SetDeaths(0, bt.SetTo, EPD(r.getValueAddr()), 0),
                    a2 << bt.SetDeaths(0, bt.SetTo, nptr, 0),
                ],
            )

            nptr << bt.NextTrigger()
            return r

        def _constget(self, i):
            r = EUDVariable()
            nptr = Forward()

            bt.RawTrigger(
                nextptr=self + 72 * i,
                actions=[
                    bt.SetDeaths(
                        self._epd + (18 * i + 344 // 4),
                        bt.SetTo,
                        EPD(r.getValueAddr()),
                        0,
                    ),
                    bt.SetDeaths(self._epd + (18 * i + 1), bt.SetTo, nptr, 0),
                ],
            )

            nptr << bt.NextTrigger()
            return r

        def set(self, i, value):
            if IsConstExpr(self) and IsConstExpr(i):
                if isUnproxyInstance(i, int):
                    ep_assert(i < size, _("EUDVArray index out of bounds."))
                if isUnproxyInstance(value, EUDVariable):
                    self._consteudset(i, value)
                else:
                    self._constset(i, value)
            else:
                self._set(i, value)

        def _set(self, i, value):
            # This function is hand-optimized

            a0, t = Forward(), Forward()
            SeqCompute(
                [
                    (EPD(a0 + 16), bt.SetTo, self._epd + (18 * i + 348 // 4)),
                    (EPD(a0 + 20), bt.SetTo, value),
                ]
            )
            t << bt.RawTrigger(actions=[a0 << bt.SetDeaths(0, bt.SetTo, 0, 0)])

        def _consteudset(self, i, value):
            VProc(value, [value.QueueAssignTo(self._epd + (18 * i + 348 // 4))])

        def _constset(self, i, value):
            bt.RawTrigger(
                actions=bt.SetDeaths(
                    self._epd + (18 * i + 348 // 4), bt.SetTo, value, 0
                )
            )

        def fill(self, values, *, assert_expected_values_len=None):
            if assert_expected_values_len:
                ep_assert(len(values) == assert_expected_values_len)

            SeqCompute(
                [
                    (self._epd + (18 * i + 348 // 4), bt.SetTo, value)
                    for i, value in enumerate(values)
                ]
            )

        def __getitem__(self, i):
            return self.get(i)

        def __setitem__(self, i, value):
            self.set(i, value)

        def __iter__(self):
            raise ut.EPError(_("Can't iterate EUDVArray"))

    return _EUDVArray



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import traceback

from .. import rawtrigger as bt
from ..allocator import Forward, ConstExpr, IsConstExpr
from ...localize import _
from ...utils import (
    FlattenList,
    EPD,
    List2Assignable,
    unProxy,
    isUnproxyInstance,
    ep_assert,
    EPError,
    ep_warn,
    RandList,
)
from .vbase import VariableBase
from .vbuf import GetCurrentVariableBuffer


isRValueStrict = False


def EP_SetRValueStrictMode(mode):
    global isRValueStrict
    isRValueStrict = mode


def _ProcessDest(dest):
    try:
        dest.checkNonRValue()
        dest = EPD(dest.getValueAddr())
    except AttributeError:
        pass
    return dest


# Unused variable don't need to be allocated.
class VariableTriggerForward(ConstExpr):
    def __init__(self, initval, mask=None):
        super().__init__(self)
        self._initval = initval
        self._mask = mask

    def Evaluate(self):
        evb = GetCurrentVariableBuffer()
        try:
            return evb._vdict[self].Evaluate()
        except KeyError:
            vt = evb.CreateVarTrigger(self, self._initval, self._mask)
            return vt.Evaluate()


class EUDVariable(VariableBase):

    """
    Full variable.
    """

    def __init__(self, initval=0):
        self._vartrigger = VariableTriggerForward(initval)
        self._varact = self._vartrigger + (8 + 320)
        self._rvalue = False

    def GetVTable(self):
        return self._vartrigger

    def getDestAddr(self):
        return self._varact + 16

    def getValueAddr(self):
        return self._varact + 20

    def __hash__(self):
        return id(self)

    # -------
    def makeL(self):
        self._rvalue = False
        return self

    def makeR(self):
        self._rvalue = True
        return self

    def checkNonRValue(self):
        if isRValueStrict and self._rvalue:
            raise EPError(_("Trying to modify value of l-value variable"))

    # -------

    def SetDest(self, dest):
        dest = _ProcessDest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.SetTo, dest)

    def AddDest(self, dest):
        dest = _ProcessDest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.Add, dest)

    def SubtractDest(self, dest):
        dest = _ProcessDest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.Subtract, dest)

    # -------

    def QueueAssignTo(self, dest):
        return [
            self.SetDest(dest),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x072D0000, 0),
        ]

    def QueueAddTo(self, dest):
        return [
            self.SetDest(dest),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x082D0000, 0),
        ]

    def QueueSubtractTo(self, dest):
        return [
            self.SetDest(dest),
            bt.SetDeaths(EPD(self._varact + 24), bt.SetTo, 0x092D0000, 0),
        ]

    # -------

    def Assign(self, other):
        self.checkNonRValue()
        SeqCompute(((self, bt.SetTo, other),))

    def __lshift__(self, other):
        self.Assign(other)
        return self

    def __iadd__(self, other):
        SeqCompute(((self, bt.Add, other),))
        return self

    def __isub__(self, other):
        if IsConstExpr(other):
            return self.__iadd__(-other)
        self << self - other
        return self

    # -------

    def __add__(self, other):
        t = EUDVariable()
        SeqCompute([(t, bt.SetTo, self), (t, bt.Add, other)])
        return t.makeR()

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        t = EUDVariable()

        # TODO: fix bug with freeing tests.
        # if IsConstExpr(other):
        #     SeqCompute(
        #         [
        #             (t, bt.SetTo, -other),
        #             (t, bt.Add, self),
        #         ]
        #     )
        # else:
        SeqCompute(
            [
                (t, bt.SetTo, 0xFFFFFFFF),
                (t, bt.Subtract, other),
                (t, bt.Add, 1),
                (t, bt.Add, self),
            ]
        )
        return t.makeR()

    def __rsub__(self, other):
        t = EUDVariable()
        SeqCompute(
            [
                (t, bt.SetTo, 0xFFFFFFFF),
                (t, bt.Subtract, self),
                (t, bt.Add, 1),
                (t, bt.Add, other),
            ]
        )
        return t.makeR()

    def __neg__(self):
        return (0 - self).makeR()

    def __invert__(self):
        t = EUDVariable()
        SeqCompute([(t, bt.SetTo, 0xFFFFFFFF), (t, bt.Subtract, self)])
        return t.makeR()

    # -------

    def __eq__(self, other):
        try:
            return self.Exactly(other)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            return (self - other).Exactly(0)

    def __ne__(self, other):
        if isinstance(other, int) and other == 0:
            return self.AtLeast(1)
        else:
            return (self - other).AtLeast(1)

    def __le__(self, other):
        try:
            return self.AtMost(other)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            t = EUDVariable()
            SeqCompute(((t, bt.SetTo, self), (t, bt.Subtract, other)))
            return t.Exactly(0)

    def __ge__(self, other):
        try:
            return self.AtLeast(other)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            t = EUDVariable()
            SeqCompute(((t, bt.SetTo, other), (t, bt.Subtract, self)))
            return t.Exactly(0)

    def __lt__(self, other):
        if isinstance(other, int) and other <= 0:
            ep_warn(_("No unsigned number can be leq than {}").format(other))
            traceback.print_stack()
            return [bt.Never()]  # No unsigned number is less than 0

        try:
            return self.AtMost(other - 1)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            t = EUDVariable()
            SeqCompute(((t, bt.SetTo, 1), (t, bt.Add, self), (t, bt.Subtract, other)))
            return t.Exactly(0)

    def __gt__(self, other):
        if isinstance(other, int) and other >= 0xFFFFFFFF:
            ep_warn(_("No unsigned number can be greater than {}").format(other))
            traceback.print_stack()
            return [bt.Never()]  # No unsigned number is less than 0

        try:
            return self.AtLeast(other + 1)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            t = EUDVariable()
            SeqCompute(((t, bt.SetTo, self), (t, bt.Subtract, other)))
            return t.AtLeast(1)

    # operator placeholders
    def __mul__(self, a):
        pass

    def __rmul__(self, a):
        pass

    def __imul__(self, a):
        pass

    def __floordiv__(self, a):
        raise NotImplementedError("")

    def __rfloordiv__(self, a):
        pass

    def __ifloordiv__(self, a):
        pass

    def __mod__(self, a):
        pass

    def __rmod__(self, a):
        pass

    def __imod__(self, a):
        pass

    def __and__(self, a):
        pass

    def __rand__(self, a):
        pass

    def __iand__(self, a):
        pass

    def __or__(self, a):
        pass

    def __ror__(self, a):
        pass

    def __ior__(self, a):
        pass

    def __xor__(self, a):
        pass

    def __rxor__(self, a):
        pass

    def __ixor__(self, a):
        pass


def IsEUDVariable(x):
    return isUnproxyInstance(x, EUDVariable)


# ---------


def VProc(v, actions):
    v = FlattenList(v)
    actions = FlattenList(actions)
    end = Forward()
    triggers = list()

    for cv, nv in zip(v, v[1:]):
        actions.append(bt.SetNextPtr(cv.GetVTable(), nv.GetVTable()))
    actions.append(bt.SetNextPtr(v[-1].GetVTable(), end))

    for i in range(0, len(actions), 64):
        trg = bt.RawTrigger(actions=actions[i : i + 64])
        triggers.append(trg)
    triggers[-1]._nextptr = v[0].GetVTable()
    end << bt.NextTrigger()

    return List2Assignable(triggers)


# From vbuffer.py
def EUDCreateVariables(varn):
    return List2Assignable([EUDVariable() for _ in range(varn)])


# -------


def _GetComputeDest(dst):
    try:
        return EPD(dst.getValueAddr())
    except AttributeError:
        return dst


def _SeqComputeSub(assignpairs, _srcdict={}):
    """
    Subset of SeqCompute with following restrictions

    - Assignment from variable should be after assignment from constant.
    - Total number of actions should be leq than 64
    """

    actionlist = []

    # Collect constant-assigning actions
    const_assigning_index = len(assignpairs)

    for i, assignpair in enumerate(assignpairs):
        dst, mdt, src = assignpair
        if IsEUDVariable(src):
            const_assigning_index = i
            break

    for dst, mdt, src in assignpairs[0:const_assigning_index]:
        dst = _GetComputeDest(dst)
        actionlist.append(bt.SetDeaths(dst, mdt, src, 0))

    # Only constant-assigning function : skip
    if const_assigning_index == len(assignpairs):
        bt.RawTrigger(actions=actionlist)
        return

    #
    # Rest is for non-constant assigning actions
    #
    nextptr = None  # nextptr for this rawtrigger
    vt_nextptr = None  # what to set for nextptr of current vtable
    last_pairs = None
    nonConstActions = list()

    def _RemoveDuplicateActions():
        if last_pairs is None:
            return
        last_src, last_dst, last_mdt = last_pairs
        try:
            prev_dst, prev_mdt, prev_nptr = _srcdict[last_src]
        except KeyError:
            pass
        else:
            lastact = nonConstActions[-1]
            queueact, setnptr = lastact
            setdst, setmdt = queueact
            if last_dst is prev_dst:
                queueact.remove(setdst)
            if last_mdt is prev_mdt:
                queueact.remove(setmdt)
            if vt_nextptr._expr is prev_nptr._expr:
                lastact.remove(setnptr)
        _srcdict[last_src] = (last_dst, last_mdt, vt_nextptr)

    for dst, mdt, src in assignpairs[const_assigning_index:]:
        dst = _GetComputeDest(dst)

        if nextptr is None:
            nextptr = src.GetVTable()
        else:
            vt_nextptr << src.GetVTable()

        _RemoveDuplicateActions()

        vt_nextptr = Forward()
        queuef = {
            bt.SetTo: EUDVariable.QueueAssignTo,
            bt.Add: EUDVariable.QueueAddTo,
            bt.Subtract: EUDVariable.QueueSubtractTo,
        }[mdt]

        nonConstActions.append(
            [queuef(src, dst), bt.SetNextPtr(src.GetVTable(), vt_nextptr)]
        )
        last_pairs = src, dst, mdt

    _RemoveDuplicateActions()
    bt.RawTrigger(nextptr=nextptr, actions=[actionlist, RandList(nonConstActions)])

    vt_nextptr << bt.NextTrigger()


def SeqCompute(assignpairs):
    # We need dependency map while writing assignment pairs
    dstvarset = set()
    srcvarset = set()

    # Record previous dst, mdt for src to optimize duplicate actions
    import inspect

    srcdictsub = inspect.signature(_SeqComputeSub).parameters["_srcdict"].default
    srcdict = {}

    # Sublist of assignments to put in _SeqComputeSub
    subassignpairs = []

    # Is we collecting constant-assigning pairs?
    constcollecting = True

    # Number of expected actions.
    actioncount = 0

    def FlushPairs():
        nonlocal constcollecting, actioncount

        if actioncount == 0:  # Already flushed before
            return

        _SeqComputeSub(subassignpairs)

        dstvarset.clear()
        srcvarset.clear()
        subassignpairs.clear()
        constcollecting = True
        actioncount = 0

    for assignpair in assignpairs:
        dst, mdt, src = assignpair
        dst = bt.EncodePlayer(unProxy(dst))
        src = unProxy(src)

        # Flush action set before preceeding
        if IsEUDVariable(src):
            if src in dstvarset:
                FlushPairs()
            elif src in srcvarset:
                FlushPairs()
            elif actioncount >= 64 - 3:
                FlushPairs()

            srcvarset.add(src)
            constcollecting = False
            actioncount += 3

            try:
                prev_dst, prev_mdt = srcdict[src]
            except KeyError:
                pass
            else:
                if dst is prev_dst:
                    actioncount -= 1
                if mdt is prev_mdt:
                    actioncount -= 1
            srcdict[src] = dst, mdt

        else:
            if not constcollecting:
                FlushPairs()
            elif actioncount >= 64 - 3:
                FlushPairs()

            actioncount += 1

        subassignpairs.append((dst, mdt, src))
        if IsEUDVariable(dst):
            dstvarset.add(dst)

    FlushPairs()
    srcdictsub.clear()


def NonSeqCompute(assignpairs):
    import itertools

    dstvarset = set()
    srcvarset = set()
    constpairs = list()
    varassigndict = dict()

    for assignpair in assignpairs:
        dst, mdt, src = assignpair
        if IsEUDVariable(dst):
            dstvarset.add(dst)
        if IsEUDVariable(src):
            srcvarset.add(src)
            try:
                varassigndict[src].append(assignpair)
            except KeyError:
                varassigndict[src] = [assignpair]
        else:
            constpairs.append(assignpair)

    if len(assignpairs) == len(constpairs):
        SeqCompute(assignpairs)
        return
    ep_assert(dstvarset.isdisjoint(srcvarset), _("dst and src have intersection"))

    varpairlists = list()
    for pairlist in varassigndict.values():
        pairlist.sort(key=lambda x: id(x[0]))
        pairlist.sort(key=lambda x: x[1]._name)
        varpairlists.append(pairlist)

    varassignpairs = list()
    for vp in itertools.zip_longest(*varpairlists):
        for varassignpair in vp:
            if varassignpair is not None:
                varassignpairs.append(varassignpair)

    SeqCompute(constpairs + varassignpairs)


def SetVariables(srclist, dstlist, mdtlist=None):
    srclist = FlattenList(srclist)
    dstlist = FlattenList(dstlist)
    ep_assert(len(srclist) == len(dstlist), _("Input/output size mismatch"))

    if mdtlist is None:
        mdtlist = [bt.SetTo] * len(srclist)

    sqa = [(src, mdt, dst) for src, dst, mdt in zip(srclist, dstlist, mdtlist)]
    SeqCompute(sqa)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .. import rawtrigger as rt
from .. import allocator as ac
from ... import utils as ut
from ...localize import _

from .eudfuncn import EUDFuncN
from .eudtypedfuncn import applyTypes
from ..variable import VProc, EUDVariable, SetVariables
from ..eudstruct import EUDStruct, EUDVArray


#
# Common argument / returns storage
#


def getArgStorage(argn, _argstorage_dict={}):
    """ Get common arguments storage for argn """
    if argn not in _argstorage_dict:
        _argstorage_dict[argn] = [EUDVariable() for _ in range(argn)]
    return _argstorage_dict[argn]


def getRetStorage(retn, _retstorage_dict={}):
    """ Get common returns storage for retn """
    if retn not in _retstorage_dict:
        _retstorage_dict[retn] = [EUDVariable() for _ in range(retn)]
    return _retstorage_dict[retn]


def fillArguments(f):
    """ Copy values from common argument storage to f._args """
    if f._argn:
        argStorage = getArgStorage(f._argn)
        for farg, arg in zip(f._fargs, argStorage):
            VProc(arg, arg.QueueAssignTo(farg))


def fillReturns(f):
    """ Copy values from f_rets to common returns storage """
    if f._retn:
        retStorage = getRetStorage(f._retn)
        for fret, ret in zip(f._frets, retStorage):
            VProc(fret, fret.QueueAssignTo(ret))


def callFuncBody(fstart, fend):
    """ Call function's body triggers """
    fcallend = ac.Forward()

    rt.RawTrigger(nextptr=fstart, actions=[rt.SetNextPtr(fend, fcallend)])

    fcallend << rt.NextTrigger()


def createIndirectCaller(f, _caller_dict={}):
    """ Create function caller using common argument/return storage """

    # Cache function in _caller_dict. If uncached,
    if f not in _caller_dict:
        rt.PushTriggerScope()
        caller_start = rt.NextTrigger()
        fillArguments(f)
        callFuncBody(f._fstart, f._fend)
        fillReturns(f)
        caller_end = rt.RawTrigger()
        rt.PopTriggerScope()

        _caller_dict[f] = (caller_start, caller_end)

    return _caller_dict[f]


# ---------------------------------


def EUDTypedFuncPtr(argtypes, rettypes):
    argn = len(argtypes)
    retn = len(rettypes)

    class PtrDataClass(EUDStruct):
        _fields_ = ["_fstart", "_fendnext_epd"]

        def __init__(self, _from=None):
            if _from is not None and isinstance(_from, EUDFuncN):
                # Statically generate with EUDFuncN
                self.checkValidFunction(_from)
                f_idcstart, f_idcend = createIndirectCaller(_from)
                super().__init__(_from=EUDVArray(2)([f_idcstart, ut.EPD(f_idcend + 4)]))
            else:
                # Cast from EUDVariable or ConstExpr
                super().__init__(_from=_from)

        @classmethod
        def cast(cls, _from):
            return cls(_from=_from)

        def checkValidFunction(self, f):
            ut.ep_assert(isinstance(f, EUDFuncN), _("{} is not an EUDFuncN").format(f))
            if not f._fstart:
                f._CreateFuncBody()

            f_argn = f._argn
            f_retn = f._retn
            ut.ep_assert(
                argn == f_argn,
                _("Function requires {} arguments (Expected {})").format(f_argn, argn),
            )
            ut.ep_assert(
                retn == f_retn,
                _("Function returns {} values (Expected {})").format(f_retn, retn),
            )

        def setFunc(self, f):
            """ Set function pointer's target to function

            :param f: Target function
            """
            try:
                self._fstart, self._fendnext_epd = f._fstart, f._fendnext_epd

            except AttributeError:
                self.checkValidFunction(f)

                # Build actions
                f_idcstart, f_idcend = createIndirectCaller(f)
                self._fstart = f_idcstart
                self._fendnext_epd = ut.EPD(f_idcend + 4)

        def __lshift__(self, rhs):
            self.setFunc(rhs)

        def __call__(self, *args):
            """ Call target function with given arguments """

            args = applyTypes(argtypes, args)

            if argn:
                argStorage = getArgStorage(argn)
                SetVariables(argStorage, args)

            # Call function
            t, a = ac.Forward(), ac.Forward()
            SetVariables(
                [ut.EPD(t + 4), ut.EPD(a + 16)], [self._fstart, self._fendnext_epd]
            )

            fcallend = ac.Forward()
            t << rt.RawTrigger(actions=[a << rt.SetNextPtr(0, fcallend)])
            fcallend << rt.NextTrigger()

            if retn:
                tmpRets = [EUDVariable() for _ in range(retn)]
                retStorage = getRetStorage(retn)
                SetVariables(tmpRets, retStorage)
                tmpRets = applyTypes(rettypes, tmpRets)
                return ut.List2Assignable(tmpRets)

    return PtrDataClass


def EUDFuncPtr(argn, retn):
    return EUDTypedFuncPtr([None] * argn, [None] * retn)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from eudplib import core as c, ctrlstru as cs


def EUDBinaryMax(cond, minv=0, maxv=0xFFFFFFFF):
    """ Find maximum x satisfying cond(x) using binary search

    :param cond: Test condition
    :param minv: Minimum value in domain
    :param maxv: Maximum value in domain

    Cond should be binary classifier, meaning that for some N
        for all x > N, cond(x) is false.
        for all x <= N, cond(x) is true
    Then EUDBinaryMin will find such N.

    .. note:: If none of the value satisfies condition, then this
        function will return maxv.
    """

    x = c.EUDVariable()
    x << minv

    if isinstance(minv, int) and isinstance(maxv, int):
        r = maxv - minv
        if r == 0:
            return minv

    else:
        r = None

    for i in range(31, -1, -1):
        if r and 2 ** i > r:
            continue

        cs.DoActions(x.AddNumber(2 ** i))
        if cs.EUDIfNot()([x <= maxv, cond(x)]):
            cs.DoActions(x.SubtractNumber(2 ** i))
        cs.EUDEndIf()

    return x


def EUDBinaryMin(cond, minv=0, maxv=0xFFFFFFFF):
    """ Find minimum x satisfying cond(x) using binary search

    :param cond: Test condition
    :param minv: Minimum value in domain
    :param maxv: Maximum value in domain

    Cond should be binary classifier, meaning that for some N
        for all x < N, cond(x) is false.
        for all x >= N, cond(x) is true
    Then EUDBinaryMin will find such N

    .. note:: If none of the value satisfies condition, then this
        function will return maxv.
    """
    x = c.EUDVariable()
    x << maxv

    if isinstance(minv, int) and isinstance(maxv, int):
        r = maxv - minv
        if r == 0:
            return minv

    else:
        r = None

    for i in range(31, -1, -1):
        if r and 2 ** i > r:
            continue

        if cs.EUDIf()([x >= 2 ** i]):
            cs.DoActions(x.SubtractNumber(2 ** i))
            if cs.EUDIfNot()([x >= minv, cond(x)]):
                cs.DoActions(x.AddNumber(2 ** i))
            cs.EUDEndIf()
        cs.EUDEndIf()

    return x



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from ..allocator import ConstExpr
from ..allocator.payload import GetObjectAddr
from eudplib import utils as ut


class EUDObject(ConstExpr):

    """
    Class for standalone object on memory

    .. note::
        Object collection occures in three steps:

        - Collection phase : collects object used in map generation. Object
        used in WritePayload method are being collected. Methods Evaluate
        and WritePayload are called during this phase.
        - Allocation phase : Object have their offset assigned. GetDataSize
        method is called on this phase, so if GetDataSize is being called,
        it means that every object required in map has been collected.
        WritePayload and GetDataSize method should behave exactly the same as
        it should on Writing phase here.
        - Writing phase : Object is written into payload.
    """

    def __init__(self):
        super().__init__(self)

    def DynamicConstructed(self):
        """ Whether function is constructed dynamically.

        Dynamically constructed EUDObject may have their dependency list
        generated during object construction. So their dependency list is
        re-examined before allocation phase.
        """
        return False

    def Evaluate(self):
        """
        What this object should be evaluated to when used in eudplib program.

        :return: Default) Memory address of this object.

        .. note::
            In overriding this method, you can use
            :func:`eudplib.GetObjectAddr`.
        """
        return GetObjectAddr(self)

    def GetDataSize(self):
        """Memory size of object."""
        raise NotImplementedError

    def CollectDependency(self, pbuffer):
        return self.WritePayload(pbuffer)

    def WritePayload(self, pbuffer):
        """Write object"""
        raise NotImplementedError



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from dataclasses import dataclass

from eudplib import core as c
from eudplib import ctrlstru as cs
from eudplib import trigtrg as tt
from eudplib import utils as ut
from eudplib.localize import _

from ..memiof import f_cunitepdread_epd, f_dwepdread_epd


@dataclass
class _UnlimiterBool:
    is_unlimiter_on: bool = False


_unlimiter = _UnlimiterBool(is_unlimiter_on=False)


def _turnUnlimiterOn():
    global _unlimiter
    _unlimiter.is_unlimiter_on = True


def _isUnlimiterOn():
    global _unlimiter
    return _unlimiter.is_unlimiter_on


def EUDLoopList(header_offset, break_offset=None):
    blockname = "listloop"
    ut.EUDCreateBlock(blockname, header_offset)

    ptr, epd = f_dwepdread_epd(ut.EPD(header_offset))

    if break_offset is not None:
        cs.EUDWhileNot()(ptr == break_offset)
    else:
        cs.EUDWhile()([ptr > 0, ptr <= 0x7FFFFFFF])

    yield ptr, epd
    cs.EUDSetContinuePoint()
    epd += 1
    f_dwepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.ep_assert(ut.EUDPopBlock(blockname)[1] is header_offset, _("listloop mismatch"))


def EUDLoopUnit():
    ut.EUDCreateBlock("unitloop", 0x628430)

    ptr, epd = f_cunitepdread_epd(ut.EPD(0x628430))

    if cs.EUDWhile()(ptr >= 1):
        yield ptr, epd
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.EUDPopBlock("unitloop")


def EUDLoopNewUnit(allowance=2):
    firstUnitPtr = ut.EPD(0x628430)
    ut.EUDCreateBlock("newunitloop", "newlo")
    tos0 = c.EUDLightVariable()
    tos0 << 0

    ptr, epd = f_cunitepdread_epd(firstUnitPtr)
    if cs.EUDWhile()(ptr >= 1):
        epd += 0xA5 // 4
        if cs.EUDIf()(c.MemoryXEPD(epd, c.AtLeast, 0x100, 0xFF00)):
            cs.DoActions(
                c.SetMemoryXEPD(epd, c.SetTo, 0, 0xFF00), epd.AddNumber(-(0xA5 // 4))
            )
            yield ptr, epd
        if cs.EUDElse()():
            cs.DoActions(tos0.AddNumber(1), epd.AddNumber(-(0xA5 // 4)))
            cs.EUDBreakIf(tos0.AtLeast(allowance))
        cs.EUDEndIf()
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.EUDPopBlock("newunitloop")


def EUDLoopUnit2():
    """EUDLoopUnit보다 약간? 빠릅니다. 유닛 리스트를 따라가지 않고
    1700개 유닛을 도는 방식으로 작동합니다.
    """

    ptr, epd = c.EUDCreateVariables(2)
    if _isUnlimiterOn():  # orderID가 [0]Die면 없는 유닛으로 판단.
        offset, cont = 0x4C // 4, c.MemoryXEPD(epd, c.Exactly, 0, 0xFF00)
    else:  # Sprite가 0이면 없는 유닛으로 판단.
        offset, cont = 0xC // 4, c.MemoryEPD(epd, c.Exactly, 0)
    cs.DoActions(ptr.SetNumber(0x59CCA8), epd.SetNumber(ut.EPD(0x59CCA8) + offset))
    if cs.EUDLoopN()(1700):
        cs.EUDContinueIf(cont)
        yield ptr, epd - offset
        cs.EUDSetContinuePoint()
        cs.DoActions(ptr.AddNumber(336), epd.AddNumber(336 // 4))
    cs.EUDEndLoopN()


def EUDLoopPlayerUnit(player):
    player = c.EncodePlayer(player)
    first_player_unit = 0x6283F8
    ut.EUDCreateBlock("playerunitloop", first_player_unit)
    ptr, epd = f_cunitepdread_epd(ut.EPD(first_player_unit) + player)

    if cs.EUDWhile()(ptr >= 1):
        yield ptr, epd
        cs.EUDSetContinuePoint()
        # /*0x06C*/ BW::CUnit*  nextPlayerUnit;
        epd += 0x6C // 4
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.EUDPopBlock("playerunitloop")


def EUDLoopBullet():
    for ptr, epd in EUDLoopList(0x64DEC4):
        yield ptr, epd


def EUDLoopSprite():
    y_epd = c.EUDVariable()
    y_epd << ut.EPD(0x629688)

    ut.EUDCreateBlock("spriteloop", "sprlo")

    if cs.EUDWhile()(y_epd < ut.EPD(0x629688) + 256):
        ptr, epd = f_dwepdread_epd(y_epd)
        if cs.EUDWhile()(ptr >= 1):
            yield ptr, epd
            cs.EUDSetContinuePoint()
            epd += 1
            f_dwepdread_epd(epd, ret=[ptr, epd])
        cs.EUDEndWhile()
        y_epd += 1
    cs.EUDEndWhile()

    ut.EUDPopBlock("spriteloop")


def EUDLoopTrigger(player):
    player = c.EncodePlayer(player)

    tbegin = tt.TrigTriggerBegin(player)
    if cs.EUDIfNot()(tbegin == 0):
        tend = tt.TrigTriggerEnd(player)
        for ptr, epd in EUDLoopList(tbegin, tend):
            yield ptr, epd
    cs.EUDEndIf()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .eperror import ep_assert
from ..localize import _


class BlockStruManager:
    def __init__(self):
        self._blockstru = []
        self._lastblockdict = {}

    def empty(self):
        return not self._blockstru


_current_bsm = BlockStruManager()  # Default one


def SetCurrentBlockStruManager(bsm):
    global _current_bsm
    old_bsm = _current_bsm
    _current_bsm = bsm
    return old_bsm


def EUDCreateBlock(name, userdata):
    _blockstru = _current_bsm._blockstru
    _lastblockdict = _current_bsm._lastblockdict

    block = (name, userdata)
    _blockstru.append(block)

    if name not in _lastblockdict:
        _lastblockdict[name] = []
    _lastblockdict[name].append(block)


def EUDGetLastBlock():
    _blockstru = _current_bsm._blockstru
    return _blockstru[-1]


def EUDGetLastBlockOfName(name):
    _lastblockdict = _current_bsm._lastblockdict

    return _lastblockdict[name][-1]


def EUDPeekBlock(name):
    lastblock = EUDGetLastBlock()
    ep_assert(lastblock[0] == name, _("Block starting/ending mismatch"))
    return lastblock


def EUDPopBlock(name):
    _blockstru = _current_bsm._blockstru
    _lastblockdict = _current_bsm._lastblockdict

    lastblock = _blockstru.pop()
    ep_assert(
        lastblock[0] == name,
        _("Block starting/ending mismatch")
        + ("\n" + _("    - Started with {}").format(lastblock[0]))
        + ("\n" + _("    - Ended with {}").format(name)),
    )
    _lastblockdict[name].pop()
    return lastblock


def EUDGetBlockList():
    return _current_bsm._blockstru



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from eudplib import utils as ut

from ... import core as c
from ...core.mapdata.stringmap import GetStringMap, GetStringSectionName
from ...localize import _
from ...trigtrg import trigtrg as tt

""" Stage 1 : works in TRIG section.
- Check if EUD is available. If not, defeat
- Initialize stage 2 & execute it
- Modifies TRIG rawtrigger's nextptr. Modification is fixed in stage 3.
"""
trglist = []


def Trigger(players=[tt.AllPlayers], *args, **kwargs):
    global trglist
    trglist.append(tt.Trigger(players=players, *args, **kwargs))


def CopyDeaths(iplayer, oplayer, copyepd=False, initvalue=None):
    if initvalue is None:
        if copyepd:
            initvalue = tt.EPD(0)
        else:
            initvalue = 0

    Trigger(actions=tt.SetDeaths(oplayer, tt.SetTo, initvalue, 0))

    for i in ut.RandList(range(2, 32)):
        if copyepd:
            subval = 2 ** (i - 2)

        else:
            subval = 2 ** i

        Trigger(
            conditions=[tt.DeathsX(iplayer, tt.AtLeast, 1, 0, 2 ** i)],
            actions=tt.SetDeaths(oplayer, tt.Add, subval, 0),
        )


def CreateVectorRelocator(chkt, payload):
    global trglist

    str_section = GetStringMap().SaveTBL()

    """
    Algorithm credit to klassical_31@naver.com

    Overall algorithm : STR <-> MRGN cycle

    MRGN:
      SetMemory(payload, Add, payload_offset // 4)

    MRGN <-> STR0  =  Trigger(actions=[
             STR1        [SetMemory(
                            mrgn.action[i].player,
                            Add,
                            prttable[j+packn] - prttable[j]
                          ) for i in range(packn)],
             STR2        SetNextPtr(mrgn, STR[k+1])
             ...      )
    """

    mrgn = 0x58DC60
    packn = 15

    #############
    # STR SECTION
    #############
    str_sled = []
    mrgn_ort = mrgn + 836
    sledheader_prt = b"\0\0\0\0" + ut.i2b4(mrgn)
    sledheader_ort = b"\0\0\0\0" + ut.i2b4(mrgn_ort)

    # apply prt
    prttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.prttable), packn):
        prts = list(map(lambda x: x // 4, payload.prttable[i : i + packn]))
        prts = prts + [-1] * (packn - len(prts))  # -1 : dummy space
        pch = [prts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(
            sledheader_prt
            + tt.Trigger(
                players=[tt.AllPlayers],
                actions=[
                    tt.SetMemory(mrgn + 328 + 32 * j + 16, tt.Add, pch[j])
                    for j in range(packn)
                ],
            )
        )

        prevoffset = prts

    # apply ort
    orttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.orttable), packn):
        orts = list(map(lambda x: x // 4, payload.orttable[i : i + packn]))
        orts = orts + [-1] * (packn - len(orts))  # -1 : dummy space
        pch = [orts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(
            sledheader_ort
            + tt.Trigger(
                players=[tt.AllPlayers],
                actions=[
                    tt.SetMemory(mrgn_ort + 328 + 32 * j + 16, tt.Add, pch[j])
                    for j in range(packn)
                ],
            )
        )

        prevoffset = orts

    # jump to ort
    str_sled.append(
        sledheader_ort
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                [
                    tt.SetMemory(
                        mrgn_ort + 328 + 32 * j + 16, tt.Add, 0xFFFFFFFF - prevoffset[j]
                    )
                    for j in range(packn)
                ],
                tt.SetMemory(mrgn_ort + 4, tt.Add, 4),  # skip garbage area
            ],
        )
    )

    # sled completed
    str_sled = b"".join(str_sled)

    str_padding_length = -len(str_section) & 3
    strsled_offset = len(str_section) + str_padding_length + 0x191943C8
    payload_offset = strsled_offset + len(str_sled) + 4
    str_section = (
        str_section + bytes(str_padding_length) + str_sled + b"\0\0\0\0" + payload.data
    )
    chkt.setsection(GetStringSectionName(), str_section)

    ##############
    # MRGN SECTION
    ##############
    mrgn_trigger = bytearray()
    mrgn_trigger_prt = (
        bytes(4)
        + ut.i2b4(prttrg_start + strsled_offset)
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                # SetDeaths actions in MRGN initially points to EPD(payload - 4)
                [
                    tt.SetMemory(payload_offset - 4, tt.Add, payload_offset // 4)
                    for _ in range(packn)
                ],
                tt.SetMemory(mrgn + 4, tt.Add, 2408),
            ],
        )
        + bytes(836)
    )
    mrgn_trigger_ort = (
        bytes(836 + 4)
        + ut.i2b4(orttrg_start + strsled_offset)
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                [
                    tt.SetMemory(payload_offset - 4, tt.Add, payload_offset)
                    for _ in range(packn)
                ],
                tt.SetMemory(mrgn_ort + 4, tt.Add, 2408),
            ],
        )
    )
    for b1, b2 in zip(mrgn_trigger_prt, mrgn_trigger_ort):
        mrgn_trigger.append(b1 ^ b2)

    oldmrgnraw = chkt.getsection("MRGN")
    mrgn_section = bytes(mrgn_trigger) + oldmrgnraw[2408 + 836 :]
    if len(mrgn_section) != 5100:
        raise RuntimeError(_("MRGN section size bug"))
    chkt.setsection("MRGN", mrgn_section)

    ##############
    # TRIG SECTION
    ##############
    trglist.clear()

    # pts[player].lasttrigger->next = value(strs) + strsled_offset

    pts = 0x51A280
    curpl = 0x6509B0

    for player in ut.RandList(range(8)):
        triggerend = ~(0x51A284 + player * 12)

        Trigger(
            players=[player],
            actions=[
                tt.SetMemory(curpl, tt.SetTo, ut.EPD(pts + 12 * player + 4)),
                tt.SetDeaths(9, tt.SetTo, triggerend, 0),  # Used in stage 2
            ],
        )

    # read pts[player].lasttrigger
    for e in ut.RandList(range(2, 32)):
        Trigger(
            conditions=tt.DeathsX(tt.CurrentPlayer, tt.AtLeast, 1, 0, 2 ** e),
            actions=tt.SetDeaths(11, tt.Add, 2 ** e, 0),
        )

    # apply to curpl
    Trigger(
        actions=[
            tt.SetDeaths(10, tt.SetTo, ut.EPD(4), 0),
            tt.SetMemory(curpl, tt.SetTo, ut.EPD(4)),
        ]
    )
    for e in ut.RandList(range(2, 32)):
        Trigger(
            conditions=tt.DeathsX(11, tt.AtLeast, 1, 0, 2 ** e),
            actions=[
                # tt.SetDeaths(11, tt.Subtract, 2 ** e, 0),
                # used for nextptr recovery in stage 3
                tt.SetDeaths(10, tt.Add, 2 ** (e - 2), 0),
                tt.SetMemory(curpl, tt.Add, 2 ** (e - 2)),
            ],
        )

    # now curpl = EPD(value(ptsprev) + 4)
    # value(EPD(value(ptsprev) + 4)) = strs + payload_offset
    # CopyDeaths(tt.EPD(strs), tt.CurrentPlayer, False, strsled_offset)
    Trigger(
        actions=[
            tt.SetDeaths(tt.CurrentPlayer, tt.SetTo, strsled_offset, 0),
            tt.SetDeaths(11, tt.SetTo, 0, 0),
        ]
    )

    # Done!
    trigdata = b"".join(trglist)

    # Stage 1 created

    # -------

    # Previous rawtrigger datas

    oldtrigraw = chkt.getsection("TRIG")
    oldtrigs = [oldtrigraw[i : i + 2400] for i in range(0, len(oldtrigraw), 2400)]
    proc_trigs = []

    # Collect only enabled triggers
    for trig in oldtrigs:
        trig = bytearray(trig)
        flag = ut.b2i4(trig, 320 + 2048)
        if flag & 8:  # Trigger already disabled
            pass

        flag ^= 8  # Disable it temporarilly. It will be re-enabled at stage 3
        trig[320 + 2048 : 320 + 2048 + 4] = ut.i2b4(flag)
        proc_trigs.append(bytes(trig))

    oldtrigraw_processed = b"".join(proc_trigs)
    chkt.setsection("TRIG", trigdata + oldtrigraw_processed)



#!/usr/bin/python
# -*- coding: utf-8 -*-


from .memiof import (
    f_epdread_epd,
    f_dwread_epd,
    f_dwepdread_epd,
    f_wread_epd,
    f_bread_epd,
    f_dwwrite_epd,
    f_wwrite_epd,
    f_bwrite_epd,
    f_posread_epd,
)

from .. import core as c, utils as ut
from ..localize import _

import functools
import traceback

rwdict = {2: (f_wread_epd, f_wwrite_epd), 1: (f_bread_epd, f_bwrite_epd)}


def _checkEPDAddr(epd):
    if ut.isUnproxyInstance(epd, c.ConstExpr):
        if epd.rlocmode == 4:
            ut.ep_warn(_("EPD check warning. Don't use raw pointer address"))
            traceback.print_stack()


@functools.lru_cache(None)
def EPDOffsetMap(ct):
    addrTable = {}
    for name, offset, size in ct:
        if size not in [4, 2, 1]:
            raise ut.EPError(_("EPDOffsetMap member size should be in 4, 2, 1"))
        if offset % size != 0:
            raise ut.EPError(_("EPDOffsetMap members should be aligned"))
        addrTable[name] = offset, size

    class OffsetMap:
        def __init__(self, epd):
            _checkEPDAddr(epd)

            self._epd = epd

        def __getattr__(self, name):
            offset, size = addrTable[name]
            offsetEPD = offset // 4
            subp = offset % 4
            if size == 4:
                return f_dwread_epd(self._epd + offsetEPD)
            else:
                return rwdict[size][0](self._epd + offsetEPD, subp)

        def getepd(self, name):
            offset, size = addrTable[name]
            ut.ep_assert(size == 4, _("Only dword can be read as epd"))
            return f_epdread_epd(self._epd + offset // 4)

        def getdwepd(self, name):
            offset, size = addrTable[name]
            ut.ep_assert(size == 4, _("Only dword can be read as epd"))
            return f_dwepdread_epd(self._epd + offset // 4)

        def getpos(self, name):
            offset, size = addrTable[name]
            ut.ep_assert(size == 4, _("Only dword can be read as position"))
            return f_posread_epd(self._epd + offset // 4)

        def __setattr__(self, name, value):
            if name == "_epd":
                super().__setattr__(name, value)
                return

            offset, size = addrTable[name]
            offsetEPD = offset // 4
            subp = offset % 4
            if size == 4:
                return f_dwwrite_epd(self._epd + offsetEPD, value)
            else:
                return rwdict[size][1](self._epd + offsetEPD, subp, value)

    return OffsetMap


EPDCUnitMap = EPDOffsetMap(
    (
        ("prev", 0x000, 4),
        ("next", 0x004, 4),
        ("hitPoints", 0x008, 4),
        ("sprite", 0x00C, 4),
        ("moveTargetPosition", 0x010, 4),
        ("moveTargetX", 0x010, 2),
        ("moveTargetY", 0x012, 2),
        ("moveTargetUnit", 0x014, 4),
        ("nextMovementWaypoint", 0x018, 4),
        ("nextTargetWaypoint", 0x01C, 4),
        ("movementFlags", 0x020, 1),
        ("currentDirection1", 0x021, 1),
        ("flingyTurnRadius", 0x022, 1),
        ("velocityDirection1", 0x023, 1),
        ("flingyID", 0x024, 2),
        ("_unknown_0x026", 0x026, 1),
        ("flingyMovementType", 0x027, 1),
        ("position", 0x028, 4),
        ("positionX", 0x028, 2),
        ("positionY", 0x02A, 2),
        ("haltX", 0x02C, 4),
        ("haltY", 0x030, 4),
        ("flingyTopSpeed", 0x034, 4),
        ("current_speed1", 0x038, 4),
        ("current_speed2", 0x03C, 4),
        ("current_speedX", 0x040, 4),
        ("current_speedY", 0x044, 4),
        ("flingyAcceleration", 0x048, 2),
        ("currentDirection2", 0x04A, 1),
        ("velocityDirection2", 0x04B, 1),
        ("playerID", 0x04C, 1),
        ("orderID", 0x04D, 1),
        ("orderState", 0x04E, 1),
        ("orderSignal", 0x04F, 1),
        ("orderUnitType", 0x050, 2),
        ("mainOrderTimer", 0x054, 1),
        ("groundWeaponCooldown", 0x055, 1),
        ("airWeaponCooldown", 0x056, 1),
        ("spellCooldown", 0x057, 1),
        ("orderTargetPosition", 0x058, 4),
        ("orderTargetX", 0x058, 2),
        ("orderTargetY", 0x05A, 2),
        ("orderTargetUnit", 0x05C, 4),
        ("shieldPoints", 0x060, 4),
        ("unitType", 0x064, 2),
        ("previousPlayerUnit", 0x068, 4),
        ("nextPlayerUnit", 0x06C, 4),
        ("subUnit", 0x070, 4),
        ("orderQueueHead", 0x074, 4),
        ("orderQueueTail", 0x078, 4),
        ("autoTargetUnit", 0x07C, 4),
        ("connectedUnit", 0x080, 4),
        ("orderQueueCount", 0x084, 1),
        ("orderQueueTimer", 0x085, 1),
        ("_unknown_0x086", 0x086, 1),
        ("attackNotifyTimer", 0x087, 1),
        ("previousUnitType", 0x088, 2),
        ("lastEventTimer", 0x08A, 1),
        ("lastEventColor", 0x08B, 1),
        ("_unused_0x08C", 0x08C, 2),
        ("rankIncrease", 0x08E, 1),
        ("killCount", 0x08F, 1),
        ("lastAttackingPlayer", 0x090, 1),
        ("secondaryOrderTimer", 0x091, 1),
        ("AIActionFlag", 0x092, 1),
        ("userActionFlags", 0x093, 1),
        ("currentButtonSet", 0x094, 2),
        ("isCloaked", 0x096, 1),
        ("movementState", 0x097, 1),
        ("buildQueue0", 0x098, 2),
        ("buildQueue1", 0x09A, 2),
        ("buildQueue2", 0x09C, 2),
        ("buildQueue3", 0x09E, 2),
        ("buildQueue4", 0x0A0, 2),
        ("energy", 0x0A2, 2),
        ("buildQueueSlot", 0x0A4, 1),
        ("uniquenessIdentifier", 0x0A5, 1),
        ("secondaryOrderID", 0x0A6, 1),
        ("buildingOverlayState", 0x0A7, 1),
        ("hpGain", 0x0A8, 2),
        ("shieldGain", 0x0AA, 2),
        ("remainingBuildTime", 0x0AC, 2),
        ("previousHP", 0x0AE, 2),
        ("loadedUnitIndex0", 0x0B0, 2),
        ("loadedUnitIndex1", 0x0B2, 2),
        ("loadedUnitIndex2", 0x0B4, 2),
        ("loadedUnitIndex3", 0x0B6, 2),
        ("loadedUnitIndex4", 0x0B8, 2),
        ("loadedUnitIndex5", 0x0BA, 2),
        ("loadedUnitIndex6", 0x0BC, 2),
        ("loadedUnitIndex7", 0x0BE, 2),
        ("spiderMineCount", 0x0C0, 1),  # 0x0C0 union, vulture
        ("pInHanger", 0x0C0, 4),
        ("pOutHanger", 0x0C4, 4),
        ("inHangerCount", 0x0C8, 1),
        ("outHangerCount", 0x0C9, 1),  # carrier
        ("parent", 0x0C0, 4),
        ("prevFighter", 0x0C4, 4),
        ("nextFighter", 0x0C8, 4),
        ("inHanger", 0x0CC, 1),  # fighter
        ("_unknown_00", 0x0C0, 4),
        ("_unknown_04", 0x0C4, 4),
        ("flagSpawnFrame", 0x0C8, 4),  # beacon
        ("addon", 0x0C0, 4),
        ("addonBuildType", 0x0C4, 2),
        ("upgradeResearchTime", 0x0C6, 2),
        ("techType", 0x0C8, 1),
        ("upgradeType", 0x0C9, 1),
        ("larvaTimer", 0x0CA, 1),
        ("landingTimer", 0x0CB, 1),
        ("creepTimer", 0x0CC, 1),
        ("upgradeLevel", 0x0CD, 1),
        ("__E", 0x0CE, 2),  # building
        ("pPowerup", 0x0C0, 4),
        ("targetResourcePosition", 0x0C4, 4),
        ("targetResourceX", 0x0C4, 2),
        ("targetResourceY", 0x0C6, 2),
        ("targetResourceUnit", 0x0C8, 4),
        ("repairResourceLossTimer", 0x0CC, 2),
        ("isCarryingSomething", 0x0CE, 1),
        ("resourceCarryCount", 0x0CF, 1),  # worker
        ("resourceCount", 0x0D0, 2),  # 0x0D0 union
        ("resourceIscript", 0x0D2, 1),
        ("gatherQueueCount", 0x0D3, 1),
        ("nextGatherer", 0x0D4, 4),
        ("resourceGroup", 0x0D8, 1),
        ("resourceBelongsToAI", 0x0D9, 1),  # resource
        ("exit", 0x0D0, 4),  # nydus
        ("nukeDot", 0x0D0, 4),  # ghost
        ("pPowerTemplate", 0x0D0, 4),  # Pylon
        ("pNuke", 0x0D0, 4),
        ("bReady", 0x0D4, 1),  # silo
        ("harvestValueLU", 0x0D0, 4),
        ("harvestValueL", 0x0D0, 2),
        ("harvestValueU", 0x0D2, 2),
        ("harvestValueRB", 0x0D4, 4),
        ("harvestValueR", 0x0D4, 2),
        ("harvestValueB", 0x0D6, 2),  # hatchery
        ("origin", 0x0D0, 4),  # powerup
        ("harvestTarget", 0x0D0, 4),
        ("prevHarvestUnit", 0x0D4, 4),
        ("nextHarvestUnit", 0x0D8, 4),  # gatherer
        ("statusFlags", 0x0DC, 4),
        ("resourceType", 0x0E0, 1),
        ("wireframeRandomizer", 0x0E1, 1),
        ("secondaryOrderState", 0x0E2, 1),
        ("recentOrderTimer", 0x0E3, 1),
        ("visibilityStatus", 0x0E4, 4),
        ("secondaryOrderPosition", 0x0E8, 4),
        ("currentBuildUnit", 0x0EC, 4),
        ("previousBurrowedUnit", 0x0F0, 4),
        ("nextBurrowedUnit", 0x0F4, 4),
        ("rallyPosition", 0x0F8, 4),  # 0x0F8 union
        ("rallyUnit", 0x0FC, 4),  # rally
        ("prevPsiProvider", 0x0F8, 4),
        ("nextPsiProvider", 0x0FC, 4),  # PsiProvider
        ("path", 0x100, 4),
        ("pathingCollisionInterval", 0x104, 1),
        ("pathingFlags", 0x105, 1),
        ("_unused_0x106", 0x106, 1),
        ("isBeingHealed", 0x107, 1),
        ("contourBoundsLU", 0x108, 4),
        ("contourBoundsL", 0x108, 2),
        ("contourBoundsU", 0x10A, 2),
        ("contourBoundsRB", 0x10C, 4),
        ("contourBoundsR", 0x10C, 2),
        ("contourBoundsB", 0x10E, 2),
        ("removeTimer", 0x110, 2),
        ("defenseMatrixDamage", 0x112, 2),
        ("defenseMatrixTimer", 0x114, 1),
        ("stimTimer", 0x115, 1),
        ("ensnareTimer", 0x116, 1),
        ("lockdownTimer", 0x117, 1),
        ("irradiateTimer", 0x118, 1),
        ("stasisTimer", 0x119, 1),
        ("plagueTimer", 0x11A, 1),
        ("stormTimer", 0x11B, 1),
        ("irradiatedBy", 0x11C, 4),
        ("irradiatePlayerID", 0x120, 1),
        ("parasiteFlags", 0x121, 1),
        ("cycleCounter", 0x122, 1),
        ("isBlind", 0x123, 1),
        ("maelstromTimer", 0x124, 1),
        ("_unused_0x125", 0x125, 1),
        ("acidSporeCount", 0x126, 1),
        ("acidSporeTime0", 0x127, 1),
        ("acidSporeTime1", 0x128, 1),
        ("acidSporeTime2", 0x129, 1),
        ("acidSporeTime3", 0x12A, 1),
        ("acidSporeTime4", 0x12B, 1),
        ("acidSporeTime5", 0x12C, 1),
        ("acidSporeTime6", 0x12D, 1),
        ("acidSporeTime7", 0x12E, 1),
        ("acidSporeTime8", 0x12F, 1),
        ("bulletBehaviour3by3AttackSequence", 0x130, 2),
        ("pAI", 0x134, 4),
        ("airStrength", 0x138, 2),
        ("groundStrength", 0x13A, 2),
        ("_repulseUnknown", 0x14C, 1),
        ("repulseAngle", 0x14D, 1),
        ("bRepMtxX", 0x14E, 1),
        ("bRepMtxY", 0x14F, 1),
    )
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import weakref
from typing import Optional
from dataclasses import dataclass

from .. import core as c, ctrlstru as cs, utils as ut
from ..core.variable import eudv as ev
from ..localize import _

from .eudarray import EUDArray


class _ObjPoolData(c.ConstExpr):
    def __init__(self, size: int, max_fieldn: int = 8):
        super().__init__(self)
        self._vdict = weakref.WeakKeyDictionary()
        self.size = size
        self.max_fieldn = max_fieldn

    def Evaluate(self):
        evb = ev.GetCurrentVariableBuffer()
        try:
            return evb._vdict[self].Evaluate()
        except KeyError:
            ret = evb.CreateMultipleVarTriggers(
                self, [0] * (self.max_fieldn * self.size)
            )
            return ret.Evaluate()


class ObjPool:
    def __init__(self, size: int, max_fieldn: int = 8):
        self.size = size
        self.max_fieldn = max_fieldn
        self.remaining = c.EUDVariable(size)

        self.baseobj = _ObjPoolData(size)
        self.data = EUDArray([72 * max_fieldn * i for i in range(size)])

    def full(self):
        return self.remaining == 0

    @c.EUDMethod
    def _alloc(self):
        """ Allocate one object from pool """
        if cs.EUDIf()(self.full()):
            c.EUDReturn(0)
        cs.EUDEndIf()

        self.remaining -= 1
        data = self.data[self.remaining]
        return data + self.baseobj

    def alloc(self, basetype, *args, **kwargs):
        ut.ep_assert(
            len(basetype._fields_) <= self.max_fieldn,
            _("Only structs less than {} fields can be allocated").format(
                self.max_fieldn
            ),
        )
        data = self._alloc()
        data = basetype.cast(data)
        data.constructor(*args, **kwargs)
        return data

    def free(self, basetype, data):
        ut.ep_assert(
            len(basetype._fields_) <= self.max_fieldn,
            _("Only structs less than {} fields can be allocated").format(
                self.max_fieldn
            ),
        )

        data = basetype.cast(data)
        data.destructor()
        self.data[self.remaining] = data - self.baseobj
        self.remaining += 1


@dataclass
class _GlobalObjPool:
    pool: Optional[ObjPool]
    max_fieldn: int = 8
    max_object_num: int = 32768


globalPool = _GlobalObjPool(pool=None, max_fieldn=8, max_object_num=32768)


def SetGlobalPoolFieldN(fieldn: int):
    global globalPool
    ut.ep_assert(globalPool.pool is None, "Global object pool is already initialized.")
    globalPool.max_fieldn = fieldn


def GetGlobalPool():
    global globalPool
    if globalPool.pool is None:
        size, max_fieldn = globalPool.max_object_num, globalPool.max_fieldn
        globalPool.pool = ObjPool(size, max_fieldn)
    return globalPool.pool



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014 trgk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from ... import utils as ut
from .. import variable as ev

from .eudfuncn import EUDFuncN
from ...localize import _


def applyTypes(typesdecl, varlist):
    """
    EUD-Typecast each variable to declared types.

    :param typesdecl: List of types. Can be set to None if you don't want
        to typecast anything. Each item of the list can also be None which is
        equivilant to EUDVariable here.
    :param varlist: List of variables.

    :returns: List of casted variables.
    """
    if typesdecl is None:
        return varlist

    ut.ep_assert(
        len(varlist) == len(typesdecl),
        _("Different number of variables({}) from type declarations({})").format(
            len(varlist), len(typesdecl)
        ),
    )

    rets = []
    for vartype, var in zip(typesdecl, varlist):
        if vartype == ev.EUDVariable or vartype is None:
            rets.append(var)
        else:
            rets.append(vartype.cast(var))

    return rets


class EUDTypedFuncN(EUDFuncN):

    """
    EUDFuncN specialization for EUDTypedFunc. This will pre-convert
    arguments to types prior to function call.
    """

    def __init__(self, argn, callerfunc, bodyfunc, argtypes, rettypes, *, traced):
        super().__init__(argn, callerfunc, bodyfunc, traced=traced)
        self._argtypes = argtypes
        self._rettypes = rettypes

    def __call__(self, *args, ret=None):
        # This layer is nessecary for function to accept non-EUDVariable object
        # as argument. For instance, EUDFuncN.
        args = applyTypes(self._argtypes, args)
        rets = super().__call__(*args, ret=ret)

        # Cast returns to rettypes before caller code.
        rets = ut.Assignable2List(rets)
        rets = applyTypes(self._rettypes, rets)
        return ut.List2Assignable(rets)


class EUDXTypedFuncN(EUDTypedFuncN):

    """
    EUDFuncN specialization for EUDX.
    """

    def __init__(
        self, argn, callerfunc, bodyfunc, argtypes, rettypes, argmasks, *, traced
    ):
        super().__init__(argn, callerfunc, bodyfunc, argtypes, rettypes, traced=traced)
        self._argmasks = argmasks

    def _CreateFuncArgs(self):
        ut.ep_assert(
            self._argn == len(self._argmasks),
            _("Different number of arguments({}) from mask declarations({})."),
        )
        if self._fargs is None:
            self._fargs = [ev.EUDXVariable(0, mask) for mask in self._argmasks]
